#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/mongo/schema.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 11.03.2018
# Last Modified Date: 26.03.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Union, Tuple, Dict, List, Type
from ampel.util.collections import check_seq_inner_type
from ampel.model.operator.AnyOf import AnyOf
from ampel.model.operator.AllOf import AllOf
from ampel.model.operator.OneOf import OneOf

VALID_TYPES = (int, str)

"""
Schema syntax example:

.. sourcecode:: python\n
	{'any_of': ["a", "b", "c"]}
	{'all_of': ["a", "b", "c"]}
	{'one_of': ["a", "b", "c"]}
	{'any_of': [{'all_of': ["a","b"]}, "c"]}
	{'any_of': [{'all_of': ["a","b"]}, {'all_of': ["a","c"]}, "d"]}


	In [1]: from ampel.mongo.schema import apply_schema, apply_excl_schema
	   ...: from ampel.util.pretty import prettyjson
	   ...: d={'$or': [{'run': 12}, {'run': 231}]}
	   ...: _=apply_schema(d, 'tag', {'any_of': [{'all_of': ["a","b"]}, "3", "1", "2"]})
	   ...: _=apply_excl_schema(d, 'channel', {'any_of': [{'all_of': ["CHAN_A","CHAN_B"]}, "CHAN_Z"]})
	   ...: print(prettyjson(d))
	{
	  "$and": [
		{"$or": [{"run": 12}, {"run": 231}]},
		{"$or": [{"tag": {"$all": ["a", "b"]}}, {"tag": {"$in": ["3", "1", "2"]}}]},
		{"channel": {"$not": {"$all": ["CHAN_A", "CHAN_B"]}}},
		{"channel": {"$ne": "CHAN_Z"}}
	  ]
	}

	In []: d={'$or': [{'run': 12}, {'run': 231}]}
	   ...: _=apply_excl_schema(d, 'channel', {'any_of': [{'all_of': ["CHAN_A","CHAN_B"]}, "CHAN_Z"]})
	   ...: _=apply_schema(d, 'tag', {'any_of': [{'all_of': ["a","b"]}, "3", "1", "2"]})
	   ...: print(prettyjson(d))
	{
	  "$and": [
		{"channel": {"$not": {"$all": ["CHAN_A", "CHAN_B"]}}},
		{"channel": {"$ne": "CHAN_Z"}},
		{"$or": [{"run": 12}, {"run": 231}]},
		{"$or": [{"tag": {"$all": ["a", "b"]}}, {"tag": {"$in": ["3", "1", "2"]}}]}
	  ]
	}

"""

def apply_schema(
	query: dict, field_name: str,
	arg: Union[int, str, dict, AllOf, AnyOf, OneOf]
) -> Dict:
	"""
	Warning: The method changes keys and values in the input dict parameter "query"

	This function translates a dict schema containing conditional (AND/OR) matching criteria
	into a dict syntax that mongodb understands. It is typically used for queries
	whose selection criteria are loaded directly from config documents.

	:param query: dict that will be later used as query criteria (can be empty).
	This dict will be updated with the additional matching criteria computed by this method.

	:param field_name: name of the target DB field

	:param arg: for convenience, the parameter can be of type int/str if only one value is to be matched.
	The dict can be nested up to one level.
	Please see :obj:`QueryMatchSchema <ampel.query.QueryMatchSchema>` docstring for details.

	:raises ValueError: if 'query' already contains the value of 'field_name' as key

	:returns: (modified/updated) dict instance referenced by query
	"""

	# Validations already performed by models
	if isinstance(arg, (AllOf, AnyOf, OneOf)):
		arg_dict = arg.dict()
	elif isinstance(arg, (int, str)):
		query[field_name] = arg
		return query
	else:
		arg_dict = arg

	if 'all_of' in arg_dict:

		# Raises error if invalid
		_check_all_of(arg_dict, VALID_TYPES)

		if len(arg_dict['all_of']) == 1:
			query[field_name] = arg_dict['all_of'][0]
		else:
			query[field_name] = {'$all': arg_dict['all_of']}

	elif 'any_of' in arg_dict:

		# Case 1: no nesting below any_of
		if check_seq_inner_type(arg_dict['any_of'], VALID_TYPES):
			if len(arg_dict['any_of']) == 1: # dumb case
				query[field_name] = arg_dict['any_of'][0]
			else:
				query[field_name] = {'$in': arg_dict['any_of']}

		# Case 2: nesting below any_of
		else:

			QueryValue = Union[int, str, Dict[str, List[Union[int, str]]]]
			QueryList = List[Dict[str, QueryValue]]
			or_list: QueryList = []
			optimize_potentially: QueryList = []

			for el in arg_dict['any_of']:

				if isinstance(el, VALID_TYPES):
					or_list.append(
						{field_name: el}
					)
					optimize_potentially.append(or_list[-1])

				elif isinstance(el, dict):

					# Raises error if invalid
					_check_all_of(el, VALID_TYPES)

					if len(el['all_of']) == 1:
						or_list.append(
							{field_name: el['all_of'][0]}
						)
						optimize_potentially.append(or_list[-1])
					else:
						or_list.append(
							{field_name: {'$all': el['all_of']}}
						)

				else:
					raise ValueError(
						f"Unsupported format: depth 2 element is not a of type {VALID_TYPES} or dict)"
						f"\nOffending value: {el}"
					)


			# ex: apply_schema(query, 'tag', {'any_of': [{'all_of': ["a","b"]}, "3", "1", "2"]})
			# optimize_potentially enables to replace:
			#   '$or': [
			#		{'tag': {'$all': ['a', 'b']}},
			#   	{'tag': '3'}, {'tag': '1'}, {'tag': '2'}
			#	]
			# with:
			#   '$or': [
			#		{'tag': {'$all': ['a', 'b']}},
			#   	{'tag': {'$in': ['3', '1', '2']}}
			#	]
			if len(optimize_potentially) > 1:
				or_list.append(
					{
						field_name: {
							'$in': [
								item
								for el in optimize_potentially
								if isinstance((item := el[field_name]), (int, str))
							]
						}
					}
				)
				for el in optimize_potentially:
					or_list.remove(el)

			if len(or_list) == 1:
				query[field_name] = or_list[0][field_name]
			else:
				if '$or' in query:
					prev_or = {'$or': query.pop('$or')}
					if '$and' in query:
						query['$and'].append(prev_or)
					else:
						query['$and'] = [prev_or]
					query['$and'].append({'$or': or_list})
				else:
					query['$or'] = or_list

	elif 'one_of' in arg_dict:
		query[field_name] = arg_dict['one_of']

	else:
		raise ValueError(
			"Invalid 'arg_dict' keys (must contain either 'any_of' or 'all_of'" +
			"\nOffending value: %s" % arg_dict
		)

	return query


def apply_excl_schema(
	query: dict, field_name: str,
	arg: Union[int, str, dict, AllOf, AnyOf, OneOf]
) -> Dict:
	"""
	Warning: The method changes keys and values in the input dict parameter "query"
	Parameters: see docstring of apply_schema
	:returns: dict instance referenced by parameter 'query' (which was updated)
	"""

	# Checks were already performed by validators
	if isinstance(arg, (AllOf, AnyOf, OneOf)):
		arg = arg.dict()
	else:
		_arg_check(arg)

	# Check if field_name criteria were set previously
	if field_name in query:
		if isinstance(query[field_name], VALID_TYPES):
			# If a previous scalar tag matching criteria was set
			# then we need rename it since query[field_name] will become a dict
			query[field_name] = {'$eq': query[field_name]}

	if isinstance(arg, VALID_TYPES):
		if field_name not in query:
			query[field_name] = {}
		query[field_name]['$ne'] = arg
		return query

	if not isinstance(arg, dict):
		raise ValueError('Illegal "arg" parameter')

	if 'all_of' in arg:

		# Raises error if invalid
		_check_all_of(arg, VALID_TYPES)

		if field_name not in query:
			query[field_name] = {}

		if len(arg['all_of']) == 1: # dumb case
			query[field_name]['$ne'] = arg['all_of'][0]
		else:
			query[field_name]['$not'] = {'$all': arg['all_of']}

	elif 'any_of' in arg:

		# Case 1: no nesting below any_of
		if check_seq_inner_type(arg['any_of'], VALID_TYPES):

			if field_name not in query:
				query[field_name] = {}

			if len(arg['any_of']) == 1: # dumb case
				query[field_name]['$ne'] = arg['any_of'][0]
			else:
				query[field_name]['$not'] = {'$all': arg['any_of']}

		else:

			QueryValue = Union[int, str, Dict[str, Union[int, str]]]
			QueryList = List[Dict[str, QueryValue]]
			and_list: QueryList = []
			optimize_potentially: QueryList = []

			for el in arg['any_of']:

				if isinstance(el, VALID_TYPES):
					and_list.append(
						{field_name: {'$ne': el}}
					)
					optimize_potentially.append(and_list[-1])

				elif isinstance(el, dict):

					# Raises error if el is invalid
					_check_all_of(el, VALID_TYPES)

					if len(el['all_of']) == 1:
						and_list.append(
							{field_name: {'$ne': el['all_of'][0]}}
						)
						optimize_potentially.append(and_list[-1])
					else:
						and_list.append(
							{
								field_name: {
									'$not': {'$all': el['all_of']} # type: ignore[dict-item]
								}
							}
						)
				else:
					raise ValueError(
						f"Unsupported format: depth 2 element is not a {VALID_TYPES} or dict)"
						f"\nOffending value: {el}"
					)

			# ex: add_from_excl_dict(query, 'tag', {'any_of': [{'all_of': ["a","b"]}, "3", "1", "2"]})
			# optimize_potentially allows to replace:
			#   '$and': [
			#		{'tag': {'$not': {'$all': ['a', 'b']}}},
			#   	{'tag': {'$ne': '3'}}, {'tag': {'$ne': '1'}}, {'tag': {'$ne': '2'}}
			#	]
			# with:
			#   '$and': [
			#		{'tag': {'$not': {'$all': ['a', 'b']}}},
			#   	{'tag': {'$nin': ['3', '1', '2']}}
			#	]
			if len(optimize_potentially) > 1:
				and_list.append(
					{
						field_name: {
							'$nin': [el[field_name]['$ne'] for el in optimize_potentially] # type: ignore
						}
					}
				)
				for el in optimize_potentially:
					and_list.remove(el)

			if len(and_list) == 1:
				query[field_name] = and_list[0][field_name]
			else:
				if '$and' in query:
					query['$and'].extend(and_list)
				else:
					query['$and'] = and_list

	elif 'one_of' in arg:
		query[field_name] = arg['one_of']

	else:
		raise ValueError(
			f"Invalid 'arg' keys (must contain either 'any_of' or 'all_of'"
			f"\nOffending value: {arg}"
		)

	return query


def _check_all_of(el, in_type: Tuple[Type, ...]) -> None:
	""" :raises: ValueError """

	if 'all_of' not in el:
		raise ValueError(
			f"Expected dict with key 'all_of'"
			f"\nOffending value: {el}"
		)

	if not check_seq_inner_type(el['all_of'], in_type):
		raise ValueError(
			f"No further nesting allowed beyond 'all_of'"
			f"\nOffending value: {el}"
		)

	if len(el) != 1:
		raise ValueError(
			f"'all_of' dict should have only one key"
			f"\nOffending value: {el}"
		)


def _arg_check(arg) -> bool:
	""" :raises: ValueError """

	if arg is None:
		raise ValueError('"arg" is None')

	if isinstance(arg, (int, str, tuple)):
		return True

	# Dict must be provided
	if not isinstance(arg, dict):
		raise ValueError(
			f"Parameter 'arg' should be a dict (is {type(arg)})"
		)

	# With only one key
	if len(arg) != 1:
		raise ValueError(
			f"Invalid 'arg' parameter. Should have only one key ('any_of' or 'all_of')"
			f"\nOffending value: {arg}"
		)

	return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/mongo/query/general.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 11.12.2019
# Last Modified Date: 17.02.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from bson.int64 import Int64
from typing import Dict, Optional, Union, Any, Literal
from ampel.types import Tag, ChannelId, StockId, StrictIterable
from ampel.model.operator.AnyOf import AnyOf
from ampel.model.operator.AllOf import AllOf
from ampel.model.operator.OneOf import OneOf
from ampel.mongo.utils import maybe_match_array
from ampel.mongo.schema import apply_schema, apply_excl_schema

type_stock_id = (int, Int64, bytes, str)


def build_general_query(
	stock: Optional[Union[StockId, StrictIterable[StockId]]] = None,
	channel: Optional[Union[ChannelId, Dict, AllOf[ChannelId], AnyOf[ChannelId], OneOf[ChannelId]]] = None,
	tag: Optional[Dict[Literal['with', 'without'], Union[Tag, Dict, AllOf[Tag], AnyOf[Tag], OneOf[Tag]]]] = None
) -> Dict[str, Any]:
	"""
	Builds a query usable with the ampel "stock", "t0" (with channel=None), "t1" and "t2" collections
	:param stock: matching multiple ids with a single query is possible
	:param channel: None (no criterium) means all channel are considered.
	:param tag: tags to be (or not to be) matched by query
	:returns: query dict with matching criteria
	:raises ValueError: apply_schema can raise ValueError in case the provided dict schema structure is unsupported
	"""

	query = {}

	if stock:
		query['stock'] = stock if isinstance(stock, type_stock_id) \
			else maybe_match_array(stock) # type: ignore[arg-type]

	if channel:
		apply_schema(query, 'channel', channel)

	if tag:

		if 'with' in tag:
			apply_schema(query, 'tag', tag['with'])

		# Order matters, parse_dict(...) must be called *after* parse_excl_dict(...)
		if 'without' in tag:
			apply_excl_schema(query, 'tag', tag['without'])

	return query



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/t3/stage/SimpleViewGenerator.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 20.04.2021
# Last Modified Date: 09.12.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Iterable, Generator
from ampel.abstract.AbsT3ReviewUnit import AbsT3ReviewUnit
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.mongo.update.MongoStockUpdater import MongoStockUpdater
from ampel.t3.stage.BaseViewGenerator import BaseViewGenerator, T, T3Send


class SimpleViewGenerator(BaseViewGenerator[T]):

	def __init__(self, unit: AbsT3ReviewUnit, buffers: Iterable[AmpelBuffer], stock_updr: MongoStockUpdater) -> None:
		super().__init__(unit_name = unit.__class__.__name__, stock_updr = stock_updr)
		self.buffers = buffers
		self.View = unit._View

	def __iter__(self) -> Generator[T, T3Send, None]:
		View = self.View
		l = self.stocks
		for ab in self.buffers:
			l.append(ab['id'])
			yield View.of(ab, freeze=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/util/collections.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 07.06.2018
# Last Modified Date: 09.12.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from itertools import islice
from collections.abc import Sequence as sequence, Iterable as iterable, Sized as sized
from typing import Any, List, Iterable, Union, Type, Tuple, Set, Generator

from ampel.types import strict_iterable, StrictIterable, T


def ampel_iter(arg: Any) -> Any:
	"""
	-> suppresses python3 treatment of str as iterable (a questionable choice)
	-> Makes None iterable
	"""
	return [arg] if isinstance(arg, (type(None), str, int, bytes, bytearray)) else arg


def chunks(seq: Iterable[T], n: int) -> Generator[List[T], None, None]:
	"""
	Yield chunks of length `n` from `seq`
	"""
	source = iter(seq)
	while True:
		if chunk := list(islice(source, n)):
			yield chunk
			if len(chunk) < n:
				break
		else:
			break


def try_reduce(arg: Any) -> Any:
	"""
	Returns element contained by sequence if sequence contains only one element.
	Example:
	try_reduce(['ab']) -> returns 'ab'
	try_reduce({'ab'}) -> returns 'ab'
	try_reduce('ab') -> returns 'ab'
	try_reduce(['a', 'b']) -> returns ['a', 'b']
	try_reduce({'a', 'b'}) -> returns {'a', 'b'}
	try_reduce(dict(a=1).keys()) -> returns 'a'
	try_reduce(dict(a=1, b=1).keys()) -> returns dict_keys(['a', 'b'])
	"""

	if isinstance(arg, sized) and len(arg) == 1:
		if isinstance(arg, sequence):
			return arg[0]
		return next(iter(arg)) # type: ignore

	return arg


def to_set(arg) -> Set:
	"""
	Reminder of python questionable logic:
	In []: set('abc')
	Out[]: {'a', 'b', 'c'}
	In []: {'abc'}
	Out[]: {'abc'}

	In []: to_set("abc")
	Out[]: {'abc'}
	In []: to_set(["abc"])
	Out[]: {'abc'}
	In []: to_set(['a','b','c'])
	Out[]: {'a', 'b', 'c'}
	In []: to_set([1,2])
	Out[]: {1, 2}
	"""
	return set(arg) if isinstance(arg, strict_iterable) else {arg}


def to_list(arg: Union[int, str, bytes, bytearray, List, Iterable]) -> List:
	"""
	raises ValueError is arg is not int, str, bytes, bytearray, list, or Iterable
	"""
	if isinstance(arg, (int, str, bytes, bytearray)):
		return [arg]
	if isinstance(arg, list):
		return arg
	if isinstance(arg, iterable):
		return list(arg)

	raise ValueError(f"Unsupported argument type ({type(arg)})")


def check_seq_inner_type(
	seq, types: Union[Type, Tuple[Type, ...]], multi_type: bool = False
) -> bool:
	"""
	check type of all elements contained in a sequence.
	*all* members of the provided sequence must match with:
		* multi_type == False: one of the provided type.
		* multi_type == True: any of the provided type.

	check_seq_inner_type((1,2), int) -> True
	check_seq_inner_type([1,2], int) -> True
	check_seq_inner_type((1,2), float) -> False
	check_seq_inner_type(('a','b'), str) -> True
	check_seq_inner_type((1,2), (int, str)) -> True
	check_seq_inner_type((1,2,'a'), (int, str)) -> False
	check_seq_inner_type((1,2,'a'), (int, str), multi_type=True) -> True

	Note:
	check_seq_inner_type('dsda', str) -> True
	check_seq_inner_type(23, int) -> False
	"""

	# Wrong input
	if not isinstance(seq, sequence) or isinstance(seq, str):
		return False

	# monotype
	if not isinstance(types, sequence):
		return all(isinstance(el, types) for el in seq)

	# different types accepted ('or' connected)
	if multi_type:
		return all(isinstance(el, types) for el in seq)

	return any(
		tuple(check_seq_inner_type(seq, _type) for _type in types)
	)


def has_nested_type(obj: StrictIterable, target_type: Type, strict: bool = True) -> bool:
	"""
	:param obj: object instance (dict/list/set/tuple)
	:param target_type: example: ReadOnlyDict/list
	:param strict: must be an instance of the provided type (subclass instances would be rejected)
	"""

	if strict:
		# pylint: disable=unidiomatic-typecheck
		if type(obj) is target_type:
			return True
	else:
		if isinstance(obj, target_type):
			return True

	if isinstance(obj, dict):
		for el in obj.values():
			if has_nested_type(el, target_type):
				return True

	elif isinstance(obj, strict_iterable):
		for el in obj:
			if has_nested_type(el, target_type):
				return True

	return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/abstract/AbsOpsUnit.py
# License           : BSD-3-Clause
# Author            : jvs
# Date              : Unspecified
# Last Modified Date: Unspecified
# Last Modified By  : jvs

from typing import Dict, Any, Optional
from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.core.ContextUnit import ContextUnit
from ampel.log.AmpelLogger import AmpelLogger
from ampel.types import Traceless

class AbsOpsUnit(AmpelABC, ContextUnit, abstract=True):
    """
    A unit for performing scheduled maintenance tasks not associated with a
    particular processing tier: collecting metrics, reporting exceptions, etc.
    """

    logger: Traceless[AmpelLogger]

    @abstractmethod
    def run(self, beacon: Optional[Dict[str, Any]] = None) -> Optional[Dict[str, Any]]:
        """
        :param beacon: the result of the previous run
        :returns:
          a BSON-serializable document summarizing the run. This will be
          supplied to the next invocation as `beacon`.
        """
        ...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/t3/supply/load/T3SimpleDataLoader.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 09.12.2019
# Last Modified Date: 29.03.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from bson.codec_options import CodecOptions
from typing import Iterable, Union, Iterator, Optional, ClassVar
from ampel.types import StockId, StrictIterable
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.abstract.AbsT3Loader import AbsT3Loader
from ampel.mongo.view.FrozenValuesDict import FrozenValuesDict


class T3SimpleDataLoader(AbsT3Loader):
	"""Load all requested documents for the selected stocks"""

	codec_options: ClassVar[Optional[CodecOptions]] = CodecOptions(document_class=FrozenValuesDict)

	def load(self,
		stock_ids: Union[StockId, Iterator[StockId], StrictIterable[StockId]]
	) -> Iterable[AmpelBuffer]:

		return self.data_loader.load(
			stock_ids = stock_ids,
			directives = self.directives,
			channel = self.channel,
			codec_options = self.codec_options,
			logger = self.logger
		)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/ingest/T1CombineCompute.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 27.05.2021
# Last Modified Date: 27.05.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Optional, Union, List
from ampel.model.UnitModel import UnitModel

class T1CombineCompute(UnitModel):
	"""
	Combine datapoints and request computation for the created t1 document
	"""

	#: Filter result codes that should trigger this T1.
	#: If not specified, T1 documents will be created in response to any passing alert.
	group: Optional[Union[int, List[int]]]

	#: T1 compute unit
	compute: UnitModel



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/config/collector/DBConfigCollector.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 16.10.2019
# Last Modified Date: 06.02.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, Any, Union
from ampel.mongo.model.AmpelDBModel import AmpelDBModel
from ampel.config.collector.ConfigCollector import ConfigCollector
from ampel.config.collector.AbsDictConfigCollector import AbsDictConfigCollector
from ampel.log import VERBOSE


class DBConfigCollector(AbsDictConfigCollector):

	def add(self,
		arg: Dict[str, Any],
		dist_name: str,
		version: Union[str, float, int],
		register_file: str
	) -> None:

		# Allow 'mongo': {'prefix': ...} in general ampel.conf
		if len(arg) == 1 and 'prefix' in arg:
			self.__setitem__('prefix', arg['prefix'])
			return

		# Allow 'mongo': {'ingest': ...} in general ampel.conf
		if len(arg) == 1 and 'ingest' in arg:
			self.__setitem__('ingest', arg['ingest'])
			return

		# At this point, arg usually contains content of files
		# contained in pyampel-core/conf/ampel-core/mongo/*

		dbs = self.get('databases')

		if not dbs:
			dbs = []
			self.__setitem__('databases', dbs)
			if 'prefix' not in self:
				self.__setitem__('prefix', 'Ampel')


		# validate model
		if self.verbose:
			self.logger.log(VERBOSE, 'Validating DB configuration')

		try:

			m = AmpelDBModel(**arg)

			if self.verbose:
				self.logger.log(VERBOSE, f'Configuration of DB collection "{m.name}" is valid')

			dbs.append(arg)

		except Exception as e:

			self.error(
				'Incorrect DB configuration ' +
				ConfigCollector.distrib_hint(dist_name, register_file),
				exc_info = e
			)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/abstract/AbsEventUnit.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 04.02.2020
# Last Modified Date: 12.12.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Any, Dict, Optional
from ampel.types import ChannelId, OneOrMany, Traceless
from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.core.ContextUnit import ContextUnit
from ampel.log.LogFlag import LogFlag


class AbsEventUnit(AmpelABC, ContextUnit, abstract=True):
	"""
	Base class for units orchestrating an ampel event.
	Ampel processes (from the ampel config) are the most common events.
	They define specifications of an event using a standardized model/schema:
	which subclass of this abstract class is to be run by a given controller
	at (a) given time(s) using a specific configuration.

	An ampel event must be registered and traceable and thus generates:
	- a new run-id
	- a new event document (in the event collection)
	- possibly Log entries (less is better)
	- new trace ids if not registered previously

	Subclassses of this class typically modify tier documents
	whereby the changes originate from results of logical unit(s)
	instantiated by this class.
	"""

	#: name of the associated process
	process_name: Traceless[str]

	#: channels associated with the process
	channel: Optional[OneOrMany[ChannelId]] = None

	#: raise exceptions instead of catching and logging
	#: if True, troubles collection will not be populated if an exception occurs
	raise_exc: bool = False

	#: logging configuration to use, from the logging section of the Ampel config
	log_profile: str = "default"

	#: flag to apply to all log records created
	base_log_flag: LogFlag = LogFlag.SCHEDULED_RUN

	#: optional additional kwargs to pass to :class:`~ampel.mongo.update.var.DBLoggingHandler.DBLoggingHandler`
	db_handler_kwargs: Optional[Dict[str, Any]] = None

	#: Some subclasses allow for non-serializable input parameter out of convenience (jupyter notebooks).
	#: For example, an AlertSupplier instance can be passed as argument of an AlertConsumer.
	#: Unless a custom serialization is implemented, this disables the hashing of input parameters
	#: and thus the generation of "trace ids".
	#: Set provenance to False to ignore trace ids generation error which will be raised otherwise.
	provenance: bool = True


	@abstractmethod(var_args=True)
	def run(self) -> Any:
		...



from typing import Any
from ampel.model.UnitModel import UnitModel
import pytest
import yaml

from ampel.model.ingest.T2Compute import T2Compute
from ampel.util.template import check_tied_units, filter_units


@pytest.fixture(scope="session")
def first_pass_config(testing_config):
    with open(testing_config) as f:
        return yaml.safe_load(f)


@pytest.mark.parametrize(
    "dependency", ["DummyPointT2Unit", "DummyStockT2Unit", "DummyStateT2Unit"]
)
def test_dependency_present(first_pass_config: dict[str, Any], dependency: str):
    units = [
        T2Compute(
            unit="DummyTiedStateT2Unit",
            config={"t2_dependency": [{"unit": dependency}]},
        ),
        T2Compute(unit=dependency),
    ]
    check_tied_units(units, first_pass_config)


def test_missing_dependency(first_pass_config: dict[str, Any]):
    units = [T2Compute(unit="DummyTiedStateT2Unit")]
    with pytest.raises(ValueError):
        check_tied_units(units, first_pass_config)


def test_misconfigured_dependency(first_pass_config: dict[str, Any]):
    units = [
        T2Compute(
            unit="DummyTiedStateT2Unit",
            config={
                "t2_dependency": [{"unit": "DummyStateT2Unit", "config": {"foo": 37}}]
            },
        ),
        T2Compute(unit="DummyStateT2Unit", config={"foo": 42}),
    ]
    with pytest.raises(ValueError):
        check_tied_units(units, first_pass_config)

@pytest.fixture
def all_units(first_pass_config: dict[str, Any]) -> list[UnitModel]:
    return [UnitModel(unit=name) for name in first_pass_config["unit"].keys()]

def test_filter_units(all_units: list[UnitModel], first_pass_config: dict[str, Any]):
    assert len(filter_units(all_units, "LogicalUnit", first_pass_config)) == 10




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/t3/T3DocBuilderModel.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 17.12.2021
# Last Modified Date: 20.12.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Optional, Literal
from ampel.types import Tag, ChannelId, OneOrMany
from ampel.base.AmpelBaseModel import AmpelBaseModel


class T3DocBuilderModel(AmpelBaseModel):
	"""
	Provides methods for handling UnitResult and generating a T3Document out of it
	"""

	channel: Optional[OneOrMany[ChannelId]] = None

	#: tag: tag(s) to add to the :class:`~ampel.content.JournalRecord.JournalRecord` of each selected stock
	extra_journal_tag: Optional[OneOrMany[Tag]] = None

	#: Record the invocation of this event in the stock journal
	update_journal: bool = True

	#: Whether t3 result should be added to t3 store once available
	propagate: bool = True

	#: Applies only for underlying T3ReviewUnits.
	#: Note that if True, a T3 document will be created even if a t3 unit returns None
	save_stock_ids: bool = False

	#: If true, value of T3Document.config will be the config dict rather than its hash
	resolve_config: bool = False

	#: Tag(s) to be added to t3 documents if applicable (if t3 unit returns something)
	tag: Optional[OneOrMany[Tag]] = None

	#: If true, value of T3Document._id will be built using the 'elements' listed below.
	#: Note that 'tag' from unit results (UnitResult.tag) if defined, will be merged
	#: with potential stager tag(s). Note also that time is always appended.
	#: ex: {_id: [DipoleJob#Task#2] [T3CosmoDipole] [2021-10-20 10:38:48.889624]}
	#: ex: {_id: [T3CosmoDipole] [TAG_UNION2] [2021-10-20 10:42:41.123263]}
	human_id: Optional[list[Literal['process', 'taskindex', 'unit', 'tag', 'config', 'run']]] = None

	#: If true, a value will be set for T3Document.datetime
	human_timestamp: bool = False

	#: Used if human_timestamp is true
	human_timestamp_format: str = "%Y-%m-%d %H:%M:%S.%f"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/t3/T3ProjectionDirective.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 02.03.2021
# Last Modified Date: 02.03.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>


from typing import Optional, Sequence
from pydantic import validator
from ampel.model.UnitModel import UnitModel
from ampel.model.StrictModel import StrictModel


class T3ProjectionDirective(StrictModel):
	"""
	Internal model used for field 'directives' of T3UnitRunner
	"""

	#: AbsT3Filter sub unit to use for down-selection of ampel buffers
	filter: Optional[UnitModel]

	#: AbsT3Projector sub unit capable of discarding selected ampel buffer attributes/fields
	project: Optional[UnitModel]

	#: t3 units (AbsT3ReviewUnit) to execute
	execute: Sequence[UnitModel]

	@validator('execute', pre=True)
	def cast_to_sequence_if_need_be(cls, v):
		if isinstance(v, dict):
			return [v]
		return v



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/t3/stage/filter/T3AmpelBufferFilter.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 14.01.2020
# Last Modified Date: 18.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

import collections
from typing import Sequence, Literal, List, Union, get_args, Iterable
from ampel.types import ChannelId
from ampel.model.operator.AnyOf import AnyOf
from ampel.model.operator.AllOf import AllOf
from ampel.model.operator.OneOf import OneOf
from ampel.struct.AmpelBuffer import AmpelBuffer, BufferKey
from ampel.model.UnitModel import UnitModel
from ampel.model.StrictModel import StrictModel
from ampel.core.UnitLoader import UnitLoader
from ampel.log.AmpelLogger import AmpelLogger
from ampel.abstract.AbsT3Filter import AbsT3Filter
from ampel.aux.filter.AbsLogicOperatorFilter import AbsLogicOperatorFilter


channel_id = get_args(ChannelId) # type: ignore[misc]


class FilterBlock(StrictModel):

	data: Union[BufferKey, Literal['journal']]
	filter: AbsLogicOperatorFilter
	include: bool = True

	def __init__(self, **kwargs):
		StrictModel.__init__(self, **kwargs)
		typ = "include" if self.include else "exclude"
		self.descr = f"{self.filter.__class__.__name__}[target={self.data}, on_match={typ}]"


class T3AmpelBufferFilter(AbsT3Filter):

	class FilterModel(StrictModel):
		data: Union[BufferKey, Literal['journal']]
		filter: UnitModel
		on_match: Literal['include', 'exclude'] = 'include'

	logger: AmpelLogger
	filters: Sequence[FilterModel] = []
	channel: Union[None, ChannelId, AllOf[ChannelId], AnyOf[ChannelId], OneOf[ChannelId]] = None

	def __init__(self, **kwargs):

		# Kulanz
		if 'filters' in kwargs and not isinstance(kwargs['filters'], collections.abc.Sequence):
			kwargs['filters'] = [kwargs['filters']]

		super().__init__(**kwargs)
		self.filter_blocks: List[FilterBlock] = []

		for f in self.filters:

			self.filter_blocks.append(
				FilterBlock(
					data = f.data,
					filter = UnitLoader.new_aux_unit(f.filter),
					include = f.on_match == "include"
				)
			)


	def filter(self, it: Iterable[AmpelBuffer]) -> Sequence[AmpelBuffer]:

		debug = self.logger.verbose > 1
		ret: List[AmpelBuffer]

		if self.channel:

			if isinstance(self.channel, channel_id):
				ret = [
					ab for ab in it
					if ab['stock']['channel'] == self.channel or # type: ignore
					self.channel in ab['stock']['channel'] # type: ignore
				]


			elif isinstance(self.channel, AnyOf):
				ret = []
				for ab in it:
					if 'stock' not in ab or not ab['stock']:
						continue
					for chan in self.channel.any_of:
						if ab['stock']['channel'] == chan or chan in ab['stock']['channel']: # type: ignore
							ret.append(ab)
							break

			elif isinstance(self.channel, AllOf):
				ret = []
				for ab in it:
					if 'stock' not in ab or not ab['stock']:
						continue
					if all([chan in ab['stock']['channel'] for chan in self.channel.all_of]): # type: ignore
						ret.append(ab)

			elif isinstance(self.channel, OneOf):
				ret = []
				for ab in it:
					if 'stock' not in ab or not ab['stock']:
						continue
					if len(ab['stock']['channel']) == 1: # type: ignore
						for chan in self.channel.one_of:
							if ab['stock']['channel'][0] == chan: # type: ignore
								ret.append(ab)
								break
			else:
				raise ValueError("Unrecognized parameter")

			# if debug and len(ret) != len(it):
			#	self.logger.debug(f"AmpelBuffer stock channel filter: in: {len(it)}, out: {len(ret)}")

		else:
			ret = list(it)

		for fb in self.filter_blocks:

			jfilter = fb.data == 'journal'

			# Using the old trick of iterating backwards to remove list element while iterating over list
			for i in range(len(ret) - 1, -1, -1):

				abuf = ret[i]

				if jfilter:
					if 'stock' not in abuf:
						raise ValueError("A StockDocument is required to filter journal entries")
					in_arr = abuf['stock']['journal'] # type: ignore
				else:
					in_arr = abuf[fb.data] # type: ignore

				fres = fb.filter.apply(in_arr)

				if debug:
					self.logger.debug(f"{fb.descr} result: in: {len(in_arr)}, out: {len(fres)}")

				if fb.include:
					if not fres: # matched elements should be include but no match
						ret.pop(i)
				elif fres: # elements were matched but on_match is exclude
					ret.pop(i)

		return ret



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/cli/ViewCommand.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 15.03.2021
# Last Modified Date: 24.03.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

import sys
from argparse import ArgumentParser
from typing import Sequence, Dict, Any, Optional, Union
from ampel.log.AmpelLogger import AmpelLogger
from ampel.log.LogFlag import LogFlag
from ampel.cli.utils import maybe_load_idmapper
from ampel.cli.ArgParserBuilder import ArgParserBuilder
from ampel.cli.AbsStockCommand import AbsStockCommand
from ampel.cli.AbsLoadCommand import AbsLoadCommand
from ampel.cli.AmpelArgumentParser import AmpelArgumentParser
from ampel.base.AuxUnitRegister import AuxUnitRegister
from ampel.model.UnitModel import UnitModel
from ampel.t3.T3Processor import T3Processor
from ampel.core.AmpelContext import AmpelContext
from ampel.model.t3.T3ProjectionDirective import T3ProjectionDirective


hlp = {
	'config': 'Path to an ampel config file (yaml/json)',
	# Optional
	'secrets': 'Path to a YAML secrets store in sops format',
	'out': 'Path to file where serialized views will be saved (printed to stdout otherwise)',
	'log-profile': 'One of: default, compact, headerless, verbose, debug',
	'id-mapper': 'Convert stock ids using the provided id mapper (ex: ZTFIdMapper)',
}

class ViewCommand(AbsStockCommand, AbsLoadCommand):
	"""
	TODO: implement
	"""

	def __init__(self):
		self.parsers = {}

	# Mandatory implementation
	def get_parser(self, sub_op: Optional[str] = None) -> Union[ArgumentParser, AmpelArgumentParser]:

		if sub_op in self.parsers:
			return self.parsers[sub_op]

		sub_ops = ['show', 'save', 'feed']
		if sub_op is None or sub_op not in sub_ops:
			return AmpelArgumentParser.build_choice_help(
				't2', sub_ops, hlp,
				description = 'Create and show or save selected ampel views.\n' +
				'These can also be provided to T3 units using the operation feed.'
			)

		builder = ArgParserBuilder("view")
		hlp.update(self.get_select_args_help())
		hlp.update(self.get_load_args_help())
		builder.add_parsers(sub_ops, hlp)

		# Required args
		builder.add_arg("required", "config", type=str)
		builder.add_arg("save.required", "out", default=True)

		# Optional args
		builder.add_arg("optional", "secrets", default=None)
		builder.add_arg('optional', 'id-mapper')
		builder.add_arg('optional', 'debug', action="store_true")
		builder.add_arg("optional", "log-profile", default="default")
		builder.add_arg("optional", "binary", action="store_true")

		# Selection args
		self.add_selection_args(builder)

		# Content args
		self.add_load_args(builder, 'Views content arguments')

		# Example
		builder.add_example("show", "view show -config ampel_conf.yaml -stock 85628462 -no-t0 -out ZTFabcdef.json")
		
		self.parsers.update(
			builder.get()
		)

		return self.parsers[sub_op]


	# Mandatory implementation
	def run(self, args: Dict[str, Any], unknown_args: Sequence[str], sub_op: Optional[str] = None) -> None:

		ctx = self.get_context(args, unknown_args, ContextClass=AmpelContext)
		maybe_load_idmapper(args)

		if sub_op == "save":
			args["out"] = open(args["out"], "wb" if args.get('binary') else 'w')

		logger = AmpelLogger.from_profile(
			ctx, 'console_debug' if args['debug'] else 'console_info',
			base_flag=LogFlag.MANUAL_RUN
		)

		conf = {
			'fd': open(args["out"], "wb" if args.get('binary') else 'w') \
				if sub_op == "save" else sys.stdout,
			'id_mapper': args["id_mapper"],
			'verbose': sub_op == "save"
		}

		t3p = T3Processor(
			context = ctx,
			process_name = "ViewCommand",
			base_log_flag = LogFlag.MANUAL_RUN,
			log_profile = 'console_debug' if args.get('debug') else 'console_info',
			update_journal = False,
			update_events = False,
			channel = args['channel'],
			supply = UnitModel(
				unit="T3DefaultBufferSupplier",
				config={
					"select": self.build_select_model(args),
					"load": self.build_load_model(args),
				},
			),
			stage = UnitModel(
				unit="T3ProjectingStager",
				config={
					"directives": [
						T3ProjectionDirective(
							project=UnitModel(
								unit="T3ChannelProjector",
								config={"channel": args['channel']}
							),
							execute=[
								UnitModel(
									unit="T3BufferBinaryExporter" if args.get('binary') \
									else "T3BufferTextExporter",
									config=conf
								)
							],
						),
					],
				},
			)
		)

		t3p.run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/t3/stage/T3AggregatingStager.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 08.12.2021
# Last Modified Date: 10.12.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from time import time
from pydantic import BaseModel
from typing import Union, Sequence, Generator, Optional, Any

from ampel.types import UBson
from ampel.t3.T3DocBuilder import T3DocBuilder
from ampel.view.T3Store import T3Store
from ampel.abstract.AbsT3Stager import AbsT3Stager
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.content.T3Document import T3Document
from ampel.content.MetaRecord import MetaRecord
from ampel.util.mappings import get_by_json_path


class TargetModel(BaseModel):
	unit: Optional[str]
	config: Union[None, int, str]
	code: Optional[int]
	field: Union[str, Sequence[str]]


class T3AggregatingStager(AbsT3Stager, T3DocBuilder):

	# Override
	paranoia: Optional[bool] = None # type: ignore
	save_stock_ids: bool = True

	split_tiers: bool = False
	t0: Union[None, TargetModel, Sequence[TargetModel]]
	t1: Union[None, TargetModel, Sequence[TargetModel]]
	t2: Union[None, TargetModel, Sequence[TargetModel]]


	def stage(self,
		gen: Generator[AmpelBuffer, None, None],
		t3s: T3Store
	) -> Optional[Generator[T3Document, None, None]]:

		t0 = [self.t0] if isinstance(self.t0, TargetModel) else self.t0
		t1 = [self.t1] if isinstance(self.t1, TargetModel) else self.t1
		t2 = [self.t2] if isinstance(self.t2, TargetModel) else self.t2

		if t0:
			t0d: dict[str, Any] = {}

		if t1:
			t1d: dict[str, Any] = {}

		if t2:
			t2d: dict[str, Any] = {}

		for ab in gen:

			sid = str(ab['id'])

			if t0:
				raise NotImplementedError("Please implement")

			if t1:
				raise NotImplementedError("Please implement")

			if t2:
				for model in t2:
					for t2doc in ab.get('t2') or []:
						if t2doc['unit'] != model.unit:
							continue
						if model.code is not None and t2doc['code'] != model.code:
							continue
						if body := self.get_t2_payload(t2doc['body'], t2doc['meta'], model.code):
							if model.field:
								for f in [model.field] if isinstance(model.field, str) else model.field:
									if (ret := get_by_json_path(body, f)):
										if sid not in t2d:
											t2d[sid] = {}
										if ret[0] == "*":
											t2d[sid].update(ret[1])
										else:
											t2d[sid][ret[0]] = ret[1]
							else:
								t2d[sid].update(ret)

		if t0 and not t1 and not t2:
			return self._craft(t0d, 't0', t3s)

		if t1 and not t0 and not t2:
			return self._craft(t1d, 't1', t3s)

		if t2 and not t1 and not t0:
			return self._craft(t2d, 't2', t3s)

		out: dict[str, Any] = {}

		if t0:
			self._upd(out, t0d, 't0')
		if t1:
			self._upd(out, t1d, 't1')
		if t2:
			self._upd(out, t2d, 't2')

		return self._craft(out, '', t3s)


	def _craft(self, d: dict[str, Any], s: str, t3s: T3Store) -> Generator[T3Document, None, None]:
		yield self.craft_t3_doc(
			self, # type: ignore
			{k: {s: v} for k, v in d.items()} if self.split_tiers else d,
			t3s,
			time(),
			[int(el) for el in d.keys()]
		)


	def _upd(self, out: dict[str, Any], d: dict[str, Any], s: str) -> None:

		if self.split_tiers:
			for k, v in d.items():
				out[k] = {s: v}
		else:
			for k, v in d.items():
				if k in out:
					out[k].update(v)
				else:
					out[k] = v


	def get_t2_payload(self,
		body: Optional[Sequence[UBson]],
		meta: Sequence[MetaRecord],
		code: Optional[int] = None
	) -> Optional[dict[str, Any]]:
		"""
		:returns: the content of the last array element of body associated with a meta code >= 0 or equals code arg.
		"""
		if not body:
			return None

		idx = len(
			[
				el for el in meta
				if el['tier'] == 2 and
				(el['code'] >= 0 if code is None else el['code'] == code)
			]
		) - 1

		if idx == -1:
			return None

		# A manual/admin $unset: {body: 1} was used to delete bad data
		if idx > len(body) - 1:
			idx = len(body) - 1

		if idx >= 0 and isinstance(body[idx], dict):
			return body[idx] # type: ignore[return-value] # remove when mypy gets smarter

		return None



from ampel.t3.supply.load.T3SimpleDataLoader import T3SimpleDataLoader
from ampel.core.AmpelContext import AmpelContext


def test_instantiate(core_config, patch_mongo, ampel_logger):
    """
    AbsT3Loader understands all the aliases in the ampel-core config
    """
    ctx = AmpelContext.load(core_config)
    aliases = ctx.config.get("alias.t3", dict)
    assert len(
        directives := T3SimpleDataLoader(
            context=ctx,
            logger=ampel_logger,
            directives=[k[1:] for k in aliases.keys()]
        ).directives
    ) == len(aliases)
    for d, value in zip(directives, aliases.values()):
        assert d.dict(exclude_defaults=True) == value



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/db/AmpelDBModel.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 19.10.2019
# Last Modified Date: 08.03.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Sequence
from ampel.mongo.model.AmpelColModel import AmpelColModel
from ampel.mongo.model.MongoClientRoleModel import MongoClientRoleModel
from ampel.model.StrictModel import StrictModel

class AmpelDBModel(StrictModel):
	name: str
	collections: Sequence[AmpelColModel]
	role: MongoClientRoleModel



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-alerts/ampel/util/register.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 24.05.2020
# Last Modified Date: 24.05.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

"""
General notes:
--------------

An ampel register file is made of three parts: signature, header and content

1) The file signature is encoded in the first 11 bytes.
The signature contains the header's length and size (3 bytes each)

2) A BSON encoded dict is saved in the next x bytes, refered to as the file's header.
The header can be updated independently of the register's content.
The size of the header is customizable during file creation.

3) The register's content is usually a compressed structure (zip, bzip or xz)
which stores information in binary format. Compression can be turned off if so whished.
(The size of a register containing 10 millions alert ids [8 bytes] and filter
return code [1 byte] can be reduced from ~90MB to ~50MB using gzip with default settings).

Properties:
----------
- Registers can be re-opened and appended
- Header content can be accessed or updated independently of the register's content.
	See methods `get_header_content` and `open_file_and_write_header` (use at your own risk)
- Header updates are fast if enough space was reserved for updates in the first place.
- Header size can be increased afterwards at the cost of having to rewrite the entire file once (function `rescale_header`)

Known class making use of this module: `ampel.core.AmpelRegister.AmpelRegister` and sub-classes
"""

import bson, json, sys
from errno import ENOENT
from struct import iter_unpack, calcsize
from os import path, strerror
from zlib import compress, decompress
if sys.version_info.minor > 8:
	from typing import TypedDict
else:
	from typing_extensions import TypedDict
from typing import BinaryIO, Optional, Dict, Any, List, Union, Tuple, Generator, Callable
from ampel.log.AmpelLogger import AmpelLogger, VERBOSE

ampel_magic_bytes = bytes([97, 109, 112, 101, 108])


class HeaderInfo(TypedDict):
	size: int     # max 16 MB
	len: int      # <= size
	payload: Dict[str, Any]


def get_outer_file_handle(
	file_path: str, write: bool = False, logger: Optional[AmpelLogger] = None,
) -> Tuple[Optional[HeaderInfo], BinaryIO]:
	"""
	If file exists, offset file handle along with decoded header dict is returned.
	Otherwise, an untouched file handle is returned (wihtout offset)  without header info.
	:raise: FileNotFoundError if write is False and file does not exist
	"""

	if file_exists := path.isfile(file_path):
		mode = 'r+b' if write else 'rb'
	else:
		if write:
			mode = 'w+b'
		else:
			raise FileNotFoundError(ENOENT, strerror(ENOENT), file_path)

	if logger:
		logger.log(VERBOSE, f"Opening {file_path} with mode {mode}")

	f: BinaryIO = open(file_path, mode) # type: ignore[assignment]
	if file_exists:
		return read_header(f, logger), f

	return None, f


def get_inner_file_handle(
	fh: Union[str, BinaryIO], write: bool = False, logger: Optional[AmpelLogger] = None
) -> BinaryIO:
	""" override if needed """

	if isinstance(fh, str):
		_, fh = get_outer_file_handle(fh, write, logger) # type: ignore

	if write:
		mode = 'ab'
		fh.seek(0, 2)
	else:
		mode = 'rb'

	if fh.name.endswith('gz'):
		from gzip import GzipFile
		if logger: logger.log(VERBOSE, f"New GzipFile from {fh.name} (mode {mode})") # noqa: E701
		return GzipFile(fileobj=fh, mode=mode) # type: ignore[return-value]

	elif fh.name.endswith('bz2'):
		from bz2 import BZ2File
		if logger: logger.log(VERBOSE, f"New BZ2File from {fh.name} (mode {mode})") # noqa: E701
		return BZ2File(fh, mode=mode) # type: ignore[return-value]

	elif fh.name.endswith('xz'):
		from lzma import LZMAFile
		if logger: logger.log(VERBOSE, f"New LZMAFile from {fh.name} (mode {mode})") # noqa: E701
		return LZMAFile(fh, mode=mode) # type: ignore[return-value]

	return fh


def get_header_size(file_path: str) -> Optional[Tuple[int, int]]:
	with open(file_path, 'rb') as f:
		return _get_header_size(f.read(11))


def _get_header_size(b_header_info: bytes) -> Optional[Tuple[int, int]]:
	"""
	:returns: (header_size, header_len) if the header signature indicates
	an ampel register file (bytes([97, 109, 112, 101, 108])), None otherwise
	:raises: ValueError if file format is not recognized
	"""

	# First 11 bytes form the file signature:
	# bytes 0->5: b'ampel' (magic)
	# bytes 5->8: header block size (not including the first 11 bytes)
	# bytes 8->11: header length
	# Note: header is padded if block size != length
	if len(b_header_info) == 0: # empty file
		return None

	if len(b_header_info) != 11 or b_header_info[:5] != ampel_magic_bytes:
		raise ValueError(
			f"Unrecognized register file (first bytes: {b_header_info})" # type: ignore[str-bytes-safe]
		)
		return None

	# Block size of the header in the file
	header_size = int.from_bytes(b_header_info[5:8], 'little')

	# Length of the header within the block
	header_len = int.from_bytes(b_header_info[8:], 'little')

	return header_size, header_len


def get_header_content(file_path: str, verbose: bool = True) -> Optional[Dict[str, Any]]:
	""" :returns: header's content as dict """

	logger = AmpelLogger.get_logger() if verbose else None
	with open(file_path, "rb") as f:
		if d := read_header(f, logger):
			return d['payload']
	return None


def read_header(
	file_handle: BinaryIO, logger: Optional[AmpelLogger] = None
) -> Optional[HeaderInfo]:
	"""
	Reads and loads header using provided file handle
	:returns: a HeaderInfo typed dict if the provided file is a non-empty ampel register file
	(starts with the bytes([97, 109, 112, 101, 108])), None otherwise
	:raises: ValueError if file format is not recognized
	"""

	size_t = _get_header_size(file_handle.read(11)) # unsigned int
	if not size_t: # empty file
		return None

	header_size = size_t[0]
	header_len = size_t[1]

	h = file_handle.read(header_size)
	if len(h) == 0: # headerless file
		return None
	elif len(h) < header_size: # something is wrong
		raise ValueError(f"{file_handle.name}: header too small (len: {len(h)})")

	header = decode_header(h, header_len)

	if logger:
		logger.log(VERBOSE, f"Header size={header_size}, len={header_len}")
		logger.log(VERBOSE, f"================== {file_handle.name} header info ==================")
		logger.log(VERBOSE, json.dumps(header, indent=4))
		logger.log(VERBOSE, "=" * (50 + len(file_handle.name)))

	return HeaderInfo(size=header_size, len=header_len, payload=header)


def decode_header(b: bytes, length: int) -> Dict[str, Any]:
	""" Strip padding, possibly decompress and bson decode header """

	hdr = b[:length]

	# Zlib compressed payload starts with b'x\x9c'
	if hdr[0] == 120 and hdr[1] == 156:
		return bson.decode(decompress(bytes(hdr)))

	return bson.decode(bytes(hdr))


def open_file_and_write_header(file_path: str, header: Dict[str, Any], verbose: bool = False) -> None:

	logger = AmpelLogger.get_logger() if verbose else None
	with open(file_path, "r+b") as f:
		if hinfo := read_header(f, logger):
			write_header(f, header=header, hsize=hinfo['size'], logger=logger)
		elif verbose:
			logger.info(f"Unable to load header info from {file_path}") # type: ignore[union-attr]


def write_header(
	file_handle: BinaryIO, header: Union[Dict[str, Any], bytes], hsize: int,
	logger: Optional[AmpelLogger] = None, flush: bool = True
) -> None:
	""" Writes (potentialy padded and compressed) file header """

	if '+' not in file_handle.mode:
		raise ValueError(f"Cannot write header, incompatible file mode {file_handle.mode}")

	if isinstance(header, dict):
		header = bson.encode(header)

	hlen = len(header)

	if logger:
		logger.log(VERBOSE, f"Writing header (block_size={hsize}, len={hlen})")

	# try to compress header if we are above the max size
	# (takes up to 80 micro-seconds on a MBP for a 4000 bytes payload)
	if hlen > hsize:
		if logger:
			logger.warn(f"Header too long ({hlen} > {hsize}), trying to compress it")
		header = compress(header) # type: ignore[arg-type]
		hlen = len(header)
		if hlen > hsize:
			raise ValueError(f"Header too long ({hlen} > {hsize})")

	file_handle.seek(0, 0)

	# file's magic
	file_handle.write(b'ampel')

	# bytes 5->8 specify header block size
	file_handle.write(int.to_bytes(hsize, 3, 'little'))

	# bytes 8->11 specify header length within header block
	file_handle.write(int.to_bytes(hlen, 3, 'little'))

	# Write bson encoded header
	i = file_handle.write(header) # type: ignore[arg-type]

	if i != hsize:
		# Padding (offsets file cursor)
		file_handle.write(bytes([0] * (hsize - i)))

	if flush:
		file_handle.flush()


def rescale_header(
	file_path: str, new_size: int, remove_old_file: bool = False,
	header: Optional[Dict[str, Any]] = None
) -> None:

	from mmap import mmap, ACCESS_WRITE
	from os import rename, remove

	logger = AmpelLogger.get_logger()
	with open(file_path, "r+b") as f1:

		if not (hinfo := read_header(f1, None if header else logger)):
			return logger.info(f"Unable to load header info from {file_path}")

		if new_size < hinfo['len']:
			raise ValueError("New size is smaller that the header length")

		# stackoverflow.com/questions/32748231/preferred-block-size-when-reading-writing-big-binary-files
		logger.info(f"Creating new {file_path}.new")
		with open(f"{file_path}.new", "w+b") as f2:

			logger.info("Writing new header")
			write_header(f2, header=header if header else hinfo['payload'], hsize=new_size, logger=logger)

			with mmap(f1.fileno(), 0) as m1:

				hdr_len_diff = new_size - hinfo['size']
				logger.info(f"Header length difference: {hdr_len_diff}")

				# OS X does not support mmap resize
				f2.seek(hdr_len_diff + len(m1) - 1)
				f2.write(b'\0')
				f2.flush()

				logger.info("Writing register content")
				with mmap(f2.fileno(), len(m1) + hdr_len_diff, access=ACCESS_WRITE) as m2:
					m2[new_size:] = m1[hinfo['size']:] # performs copy

	if remove_old_file:
		logger.info(f"Removing {file_path}")
		remove(file_path)
	else:
		logger.info(f"Renaming {file_path} -> {file_path}.old")
		rename(file_path, f"{file_path}.old")

	logger.info(f"Renaming {file_path}.new -> {file_path}")
	rename(f"{file_path}.new", file_path)


def _quick_load(
	f: Union[BinaryIO, str], logger: Optional[AmpelLogger] = None
) -> Tuple[Optional[HeaderInfo], BinaryIO, Optional[BinaryIO]]:
	""" :raise: FileNotFoundError if provided file path (str) does not exist """

	if isinstance(f, str):
		hinfo, f = get_outer_file_handle(f, logger=logger)
	else:
		f.seek(0, 0)
		hinfo = read_header(f, logger)

	if hinfo is None:
		return None, f, None

	return hinfo, f, get_inner_file_handle(f, logger=logger)


def reg_iter(
	f: Union[BinaryIO, str], read_multiplier: int = 100000, verbose: bool = True
) -> Generator[Tuple[int, ...], None, None]:
	"""
	Iterates through unpacked blocks of a register
	:param f: file path (str) or file handle (which will not be closed)
	:param read_multiplier: (<struct size> * `read_multiplier`) bytes will be read at once iteratively from the provided file handle
	:returns: yields Tuple[<elements defined in struct>] ex: Tuple[<alert id>, <filter return code>]
	"""

	logger = AmpelLogger.get_logger() if verbose else None
	hinfo, ofh, ifh = _quick_load(f, AmpelLogger.get_logger() if verbose else None)
	if not hinfo or not ifh:
		if logger: logger.info("reg_iter() cannot continue as header info are missing") # noqa
		return


	struct = hinfo['payload']['struct']
	buf_len = calcsize(struct) * read_multiplier
	r = ifh.read

	while b := r(buf_len):
		for el in iter_unpack(struct, b):
			yield el

	ifh.close()
	if isinstance(f, str) and not ofh.closed:
		ofh.close()


def find(
	f: Union[BinaryIO, str],
	offset: int,
	match_int: Optional[Union[int, List[int]]] = None,
	int_bytes_len: Optional[int] = None,
	match_bytes: Optional[Union[bytes, List[bytes]]] = None,
	header_hint: Optional[str] = None,
	header_hint_callback: Optional[Callable] = None,
	read_multiplier: int = 100000, verbose: bool = True
) -> Optional[List[Tuple[int, ...]]]:
	"""
	:param f: file path (str) or file handle (which will be closed)
	:param read_multiplier: (<struct size> * `read_multiplier`) bytes will be read at once iteratively from the provided file handle
	:param match_bytes: bytes to match in the underlying stream (ex: to match an integer x < 255, use: int.to_bytes(x, 1, 'little'))
	:param offset: position of the field of interest within each block. (ex: if the block are made of '<QB' and you want
	to match the last byte, then offset should be set to 8)
	:returns: list of matching blocks

	Example:
	Let's consider MinimalRegister, it saves the alert id as a 8 bytes integer and the filter result as single byte.
	That forms blocks of 9 bytes each. The struct.pack argument is '<QB' -> (little endian) long long and one byte.
	If we wanted to search for any register entry with filter result "10", we would call:
	find(<file>, match_bytes=int.to_bytes(10, 1, 'little'), offset=8)
	(Since the filter result is the 9th byte of each block, we did set offset equals to 8).

	# Find alert by id using the general method find
	In []: find('/Users/hu/Documents/ZTF/test/aa/aa.bin.gz', match_bytes=int.to_bytes(1242886, 8, 'little'), offset=0)
	Out[]: [(1242886, 16)]

	# Find register entries with alert filter results equals 176 or 16
	In []: find('/Users/hu/Documents/ZTF/test/aa/aa.bin.gz', match_bytes=[
		int.to_bytes(176, 1, 'little'), int.to_bytes(16, 1, 'little')], offset=8)
	Out:[]: [(9659062, 16), (7559029, 176)]
	"""

	logger = AmpelLogger.get_logger() if verbose else None
	hinfo, ofh, ifh = _quick_load(f, AmpelLogger.get_logger() if verbose else None)
	if not hinfo or not ifh:
		if logger: logger.info("find() cannot continue as header info are missing") # noqa
		return None

	if match_int:

		if not int_bytes_len:
			# In the future, we might extract automcatically int_bytes_len from 'struct' (header)
			# using the value of parameter 'offset' and method struct.calcsize()
			raise ValueError("Parameter 'int_bytes_len' must be defined with 'match_int'")

		if header_hint and header_hint in hinfo['payload']:

			d = hinfo['payload'][header_hint]
			if isinstance(match_int, int):

				if not d['min'] <= match_int <= d['max']:
					if logger: logger.info(f"Fast header check: no match for {match_int}") # noqa: E701
					return None
			else:

				new_match_int = [
					el for el in match_int
					if (d['min'] <= match_int <= d['max'])
				]

				if not new_match_int:
					if logger: logger.info("Fast header check: no match for provided integers") # noqa: E701
					return None

				if len(new_match_int) != len(match_int):
					match_int = new_match_int
					if logger: logger.info(f"Fast header check: find reduced to: {match_int}") # noqa: E701
				elif logger:
					logger.info("Fast header check: target ids are all eligible")

		if header_hint_callback:
			match_int = header_hint_callback(hinfo['payload'], match_int, logger)
			if not match_int:
				if logger: logger.info("Fast header callback check: no match") # noqa: E701
				return None

	elif match_bytes:
		pass
	else:
		raise ValueError("Please provide either parameter 'match_bytes' or 'match_int' and 'int_bytes_len'")

	if match_int:
		match_bytes = convert_to_bytes(match_int, int_bytes_len) # type: ignore[arg-type]

	func = find_one if isinstance(match_bytes, bytes) else find_many
	struct = hinfo['payload']['struct']
	block_len = calcsize(struct)
	matches = func(ifh, block_len * read_multiplier, block_len, offset, match_bytes) # type: ignore[operator]

	ifh.close()
	if isinstance(f, str) and not ofh.closed:
		ofh.close()

	if matches:
		return list(iter_unpack(struct, b''.join(matches)))

	return None


def find_one(
	f: BinaryIO, read_len: int, block_len: int, offset: int, match_bytes: bytes
) -> List[bytes]:
	"""
	:param f: file handle
	:param read_len: number of bytes read at once iteratively from the provided file handle
	:param offset: position of the field of interest within each block. (ex: if register is made of '<QB' blocks
	and you want to match the last byte, then offset should be set to 8)
	:param match_bytes: bytes to match in the underlying stream
	:returns: list of matching bytes blocks
	"""

	ret = []
	while buf := f.read(read_len):
		start = 0
		while (i := buf.find(match_bytes, start)) != -1:
			start = i + 1
			pos = i - offset # block start position
			if pos % block_len == 0: # match concerns another field
				ret.append(buf[pos: pos + block_len])
	return ret


def find_many(
	f: BinaryIO, read_len: int, block_len: int, offset: int, match_bytes: List[bytes],
) -> List[bytes]:
	"""
	:param f: file handle
	:param read_len: number of bytes read at once iteratively from the provided file handle
	:param offset: position of the field of interest within each block. (ex: if register is made of '<QB' blocks
	and you want to match the last byte, then offset should be set to 8)
	:param match_bytes: list of bytes to match in the underlying stream
	:returns: list of matching bytes blocks
	"""

	ret = []
	while buf := f.read(read_len):
		for mb in match_bytes:
			start = 0
			while (i := buf.find(mb, start)) != -1:
				start = i + 1
				pos = i - offset # block start position
				if pos % block_len == 0: # match concerns another field
					ret.append(buf[pos: pos + block_len])
	return ret


def convert_to_bytes(arg: Union[int, List[int]], bytes_len: int) -> Union[bytes, List[bytes]]:
	if isinstance(arg, int):
		return int.to_bytes(arg, bytes_len, 'little')
	return [int.to_bytes(el, bytes_len, 'little') for el in arg]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/ingest/T2Compute.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 10.03.2020
# Last Modified Date: 28.09.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import List, Union, Optional, Generic
from ampel.types import T
from ampel.model.UnitModel import UnitModel
from ampel.model.DPSelection import DPSelection

class T2Compute(UnitModel[T], Generic[T]):

	#: Ingester options
	ingest: Optional[Union[str, DPSelection]]

	#: Filter result codes that should trigger this T2. If not specified, T2
	#: documents will be created in response to any passing alert.
	group: Union[int, List[int]] = []



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/t3/supply/complement/T3ExtJournalAppender.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 17.06.2020
# Last Modified Date: 14.12.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from pymongo import MongoClient
from typing import Iterable, Optional, Union, List
from ampel.types import StockId
from ampel.aux.filter.SimpleDictArrayFilter import SimpleDictArrayFilter
from ampel.content.JournalRecord import JournalRecord
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.abstract.AbsBufferComplement import AbsBufferComplement
from ampel.model.aux.FilterCriterion import FilterCriterion
from ampel.model.operator.AllOf import AllOf
from ampel.model.operator.FlatAnyOf import FlatAnyOf
from ampel.view.T3Store import T3Store


class T3ExtJournalAppender(AbsBufferComplement):
	"""
	Import journal entries from a 'foreign' database, e.g. one created
	by a previous version of Ampel.
	"""

	mongo_resource: str = "resource.ext_mongo"
	db_name: str = "Ampel_data"
	sort: bool = True
	reverse: bool = True
	filter_config: Optional[
		Union[FilterCriterion, FlatAnyOf[FilterCriterion], AllOf[FilterCriterion]]
	] = None


	def __init__(self, **kwargs) -> None:

		super.__init__(**kwargs)

		if self.filter_config:
			self.journal_filter: SimpleDictArrayFilter[JournalRecord] = SimpleDictArrayFilter(filters=self.filter_config)

		self.col = MongoClient(self.context.config.get(f'resource.{self.mongo_resource}')) \
			.get_database(self.db_name)\
			.get_collection("stock")


	def get_ext_journal(self, stock_id: StockId) -> Optional[List[JournalRecord]]:

		if ext_stock := next(self.col.find({'_id': stock_id}), None):
			if self.journal_filter:
				return self.journal_filter.apply(ext_stock['journal'])
			return ext_stock['journal']
		return None


	def complement(self, it: Iterable[AmpelBuffer], t3s: T3Store) -> None:

		for albuf in it:

			if 'stock' in albuf and isinstance(albuf['stock'], dict):

				if entries := self.get_ext_journal(albuf['stock']['stock']):

					entries.extend(albuf['stock']['journal'])

					if self.sort:
						dict.__setitem__(
							albuf['stock'], 'journal', sorted( # type: ignore[index]
								entries, key=lambda x: x['ts'], reverse=self.reverse
							)
						)
			else:
				print(albuf)
				raise ValueError("No stock information available")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/core/EventHandler.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 26.09.2018
# Last Modified Date: 17.12.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from time import time
from bson import ObjectId
from typing import Any, Optional, Literal, TYPE_CHECKING
from ampel.content.EventDocument import EventDocument
from ampel.log.AmpelLogger import AmpelLogger
from ampel.log.AmpelLoggingError import AmpelLoggingError
from ampel.log.utils import report_exception

if TYPE_CHECKING:
	from ampel.core.AmpelDB import AmpelDB


# TODO: (much later) remove explicit dependency on pymongo
class EventHandler:
	"""
	Handles the creation and publication of event documents into the event database
	"""

	def __init__(self,
		process_name: str,
		ampel_db: 'AmpelDB',
		tier: Literal[-1, 0, 1, 2, 3],
		run_id: int,
		col_name: str = "events",
		extra: Optional[dict[str, Any]] = None,
		raise_exc = False,
		dry_run: bool = False
	):
		"""
		:param col_name: name of db collection to use (default 'events').
		"""

		self.process_name = process_name
		self.db = ampel_db
		self.raise_exc = raise_exc
		self.dry_run = dry_run
		self.run_id = run_id
		doc = EventDocument(process=process_name, tier=tier)

		if run_id:
			doc['run'] = run_id

		if extra:
			doc |= extra # type: ignore[assignment]

		self.dkeys = doc.keys()
		self.extra: Optional[dict[str, Any]] = None

		if dry_run:
			self.col = None
			self.ins_id = ObjectId()
		else:
			self.col = ampel_db.get_collection(col_name)
			self.ins_id = self.col.insert_one(doc).inserted_id


	def get_run_id(self) -> int:
		return self.run_id


	def add_extra(self, overwrite: bool = False, **extra) -> None:

		if self.extra is None:
			self.extra = extra
			return

		for k, v in extra.items():
			if k in self.extra and not overwrite:
				continue
			self.extra[k] = v


	def handle_error(self, e: Exception, logger: AmpelLogger) -> None:

		self.add_extra(overwrite=True, success=False)

		if self.raise_exc:
			raise e

		report_exception(
			self.db, logger, exc=e,
			info={'process': self.process_name}
		)


	def update(self, logger: AmpelLogger, save_duration: bool = True, **kwargs) -> None:
		""" :raises: AmpelLoggingError """

		upd: dict[str, Any] = {}

		if self.extra:
			for k, v in self.extra.items():
				if k in self.dkeys:
					logger.error(f"Cannot overwrite already existing event value for key {k}")
					continue
				upd[k] = v

		for k, v in kwargs.items():
			if k in self.dkeys:
				logger.error(f"Cannot overwrite already existing event value for key {k}")
				continue
			upd[k] = v

		if save_duration:
			upd['duration'] = round(time() - self.ins_id.generation_time.timestamp(), 3)
		elif not upd:
			return

		if not self.dry_run:
			res = self.col.update_one({'_id': self.ins_id}, {'$set': upd}) # type: ignore[union-attr]

			if res.modified_count == 0 and res.upserted_id is None:
				raise AmpelLoggingError(
					"Events collection update failed (%s)" % {
						'mongoUpdateResult': res.raw_result,
						'process': self.process_name
					}
				)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/aux/ComboDictModifier.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 21.02.2020
# Last Modified Date: 21.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, Sequence, Union, Literal, List, Optional, Callable, Container, Any, Set
from ampel.abstract.AbsApplicable import AbsApplicable
from ampel.base.AuxUnitRegister import AuxUnitRegister
from ampel.model.StrictModel import StrictModel
from ampel.log import AmpelLogger, VERBOSE
from ampel.util.collections import to_set
from ampel.view.ReadOnlyDict import ReadOnlyDict


class ComboDictModifier(AbsApplicable):
	"""
	Note: cannot make different modifications of the same ast at different depth
	"""

	class DeleteModel(StrictModel):
		"""
		ex: for a given dict instance *d*, the model:
			{'op': 'delete', 'key': ['a', 'b.c']}
		will be equivalent to:
			del d['a']
			del d['b']['c']
		"""
		op: Literal['delete']
		key: Sequence[str]

		def __init__(self, **kwargs):
			if 'key' in kwargs and isinstance(kwargs['key'], str):
				kwargs['key'] = [kwargs['key']]
			StrictModel.__init__(self, **kwargs)


	class KeepOnlyModel(StrictModel):
		"""
		ex: for a given dict instance *d*, the model:
		{'op': 'keep_only', 'key': 'a.b', 'keep': ['y', 'z']}
		will keep only the keys 'y' and 'z' of d['a']['b']
		"""
		op: Literal['keep_only']
		key: Optional[str]
		keep: Union[int, str, Sequence[Union[int, str]]]


	class ClassModifyModel(StrictModel):
		"""
		ex: for a given dict instance *d*, the model:
			{'op': 'modify', 'key': 'a.b', 'unit': 'SetIntersector', config: {'value': [1, 2]}}
		will replace the value of d['a']['b'] with the intersection of the value of d['a']['b']
		and the value provided by config.value
		"""
		op: Literal['modify']
		key: str
		unit: str
		config: Dict[str, Any]


	class FuncModifyModel(StrictModel):
		op: Literal['modify']
		key: str
		func: Callable[[Any], Any]


	logger: AmpelLogger
	modifications: Sequence[Union[DeleteModel, KeepOnlyModel, ClassModifyModel, FuncModifyModel]]

	# Whether fields can be directly altered/modified or not
	# If not, new dict instances must be created to modify existing dicts
	unalterable: bool = False

	# Whether to cast structures into immutables objects after modification
	freeze: bool = True


	def __init__(self, **kwargs) -> None:
		"""
		:param unalterable: whether fields can be directly altered/modified or not
		If not, new dict instances must be created to modify existing dicts

		:param freeze: whether to cast structures into immutables objects after modification

		:param modifications: sequence of modifications to be performed on the dict.
		Note that different modifications of the same ast at different depth is not possible
		For more information, see docstrings of internal static models:
		DeleteModel, KeepOnlyModel, ClassModifyModel (inherits AbsApplicable), FuncModifyModel
		"""

		super().__init__(**kwargs)

		# Dict ops on root dict keys
		self._mods: Dict[str, List[Callable[[Any], Any]]] = {}
		self._dels: Set[str] = set()
		self._root_ko: Container[str] = set()
		self._depth1_ko: Dict[str, Container[str]] = {}

		# Mypy does not yet support recursive types
		# RecursiveModType = Dict[str, Union[Callable[[Dict], Dict], "RecursiveModType"]]
		# RecursiveDelType = Dict[str, Union[Dict[str, "RecursiveDelType"]]]

		self._nested_dels: Dict[str, Any] = {}
		self._nested_ko: Dict[str, Any] = {}
		self._nested_mods: Dict[str, Any] = {}

		for f in self.modifications:

			if isinstance(f, self.DeleteModel):

				for k in f.key:
					if "." not in k: # no nesting
						self._dels.add(k)
					else:
						# creates: {'a':{}, 'b': {'c': {}, 'd': {'e': {}}}}
						d = self._nested_dels
						for kk in k.split("."):
							if kk not in d:
								d[kk] = {}
							d = d[kk]

			elif isinstance(f, self.KeepOnlyModel):

				# "Root" keep only
				if not f.key:
					self._root_ko.update(
						to_set(f.keep)
					)
					continue

				# Non-nested key
				if "." not in f.key:
					if f.key in self._depth1_ko:
						raise ValueError("'keep only' dict operations must be atomar")
					self._depth1_ko[f.key] = to_set(f.keep)
					continue

				# creates: {'a': set(['foo', 'bar', ...]), 'b': {'c': set(['foo', 'bar', ...]), ...}}
				keys = f.key.split(".")
				d = self._nested_ko
				for k in keys[:-1]:
					if kk not in d:
						d[kk] = {}
					d = d[kk]
				d[keys[-1]] = to_set(f.keep)

			elif isinstance(f, (self.ClassModifyModel, self.FuncModifyModel)):

				if isinstance(f, self.ClassModifyModel):
					unit = AuxUnitRegister.new_unit(
						class_name = f.unit,
						sub_type = AbsApplicable,
						logger = self.logger,
						**f.config
					)
					func = unit.apply
				else:
					func = f.func # type: ignore

				# Non-nested key
				if "." not in f.key:
					if f.key not in self._mods:
						self._mods[f.key] = []
					self._mods[f.key].append(func)
					continue

				# creates: {'a': [<AbsApplicable>, ..], 'b': {'c': [<AbsApplicable>], ...}}
				keys = f.key.split(".")
				d = self._nested_mods
				for k in keys[:-1]:
					if kk not in d:
						d[kk] = {}
					d = d[kk]
				if keys[-1] not in d:
					d[keys[-1]] = []
				d[keys[-1]].append(func)

			else:
				raise ValueError("Unrecognized config model")

		# Optimization
		if self._nested_mods and self._mods:

			for k in self._mods:
				if k in self._nested_mods:
					if isinstance(self._nested_mods[k], list):
						self._nested_mods[k].append(self._mods[k])
					else:
						raise ValueError("Cannot perform multiple modifications on the same ast")
				else:
					self._nested_mods[k] = [self._mods[k]]
			self._mods = {}


		ops: Dict[str, Any] = {
			# Ops handling non-nested keys
			"apply_root_delete": self._dels,
			"apply_root_modify": self._mods,
			"apply_root_keep_only": self._root_ko,
			"apply_depth1_keep_only": self._depth1_ko,

			# Ops handling nested keys
			"apply_delete": self._nested_dels,
			"apply_modify": self._nested_mods,
			"apply_keep_only": self._nested_ko,
		}

		if self.logger.verbose:
			self.logger.log(VERBOSE, "Following dict operations will be applied")
			for k, v in ops.items():
				if v:
					self.logger.log(VERBOSE, f" -> {k} ({len(v)})")

		self.ops = [getattr(self, k) for k, v in ops.items() if v]

		# Optimization
		different_ops = sum([1 for instructions in ops.values() if instructions])
		if different_ops == 1:
			for k, v in ops.items():
				if v:
					self.apply = getattr(self, k) # type: ignore
					break


	# =========================================
	# Modifications of non-nested dict elements
	# =========================================

	# Delete dict elements at root level
	def apply_root_delete(self, d: Dict) -> Dict:
		""" deletes given keys of dict instance """

		if self.unalterable:
			d = {k: v for k, v in d.items() if k not in self._dels}
		else:
			di = dict.__delitem__
			for k in self._dels:
				if k in d:
					di(d, k)

		return ReadOnlyDict(d) if self.freeze else d


	# "Keep only" specified dict keys at root level
	def apply_root_keep_only(self, d: Dict) -> Dict:
		"""
		Deletes all but certain dict keys of a dict (at root level).
		For example, applying model: {'op': 'delete', 'keep': ['a']}
		to dict: {'a': {'b', 'c'}, 'd': {'e': {'f'}}}
		will result in: {'a': {'b', 'c'}}
		"""
		dd = {k: v for k, v in d.items() if k in self._root_ko}
		return ReadOnlyDict(dd) if self.freeze else dd


	# "Keep only" for keys with depth equals 1
	def apply_depth1_keep_only(self, d: Dict) -> Dict:
		"""
		Deletes all but certain dict keys for a given given root keys.
		For example, applying model:
			{'op': 'delete', 'key': 'a', 'keep': ['b']}
		to dict:
			{'a': {'b', 'c'}, 'd': {'e': {'f'}}}
		will result in:
			{'a': {'b'}, 'd': {'e': {'f'}}}
		Note 1: this method does not support nesting, meaning that the
		value associated with 'key' in the model above cannot contain dots
		Note 2: Model field 'key' must target a dict instance (not a list or set)
		"""
		if self.unalterable:
			d = d.copy()

		dsi = dict.__setitem__
		for k in self._depth1_ko:
			dsi(d, k, {kk: d[k][kk] for kk in d[k] if kk in self._depth1_ko[k]})

		return ReadOnlyDict(d) if self.freeze else d


	# Modifications of dict elements at root level
	def apply_root_modify(self, d: Dict) -> Dict:
		""" Modifies root elements of dict instances """

		if self.unalterable:
			d = d.copy()

		dsi = dict.__setitem__
		for k, mods in self._mods.items():
			if k in d:
				for mod in mods:
					dsi(d, k, mod(d[k]))

		return ReadOnlyDict(d) if self.freeze else d


	# =====================================
	# Modifications of nested dict elements
	# =====================================

	# Delete nested dict elements of any depth
	def apply_delete(self, d: Dict, deletions: Optional[Dict] = None) -> Dict:
		"""
		Method apply_delete:
		:param deletions: if None, self._nested_dels is used (set by constructor)
		deletions example: {'a':{}, 'b': {'c': {}, 'd': {'e': {}}}}
		-> empty dicts mark entries to be deleted
		"""
		if not deletions:
			deletions = self._nested_dels

		if self.unalterable:
			# suppreses a
			d = {k: d[k] for k in d if k not in deletions or deletions.get(k)}

		ddi = dict.__delitem__
		for k in deletions:
			if k in d:
				if not deletions[k]: # suppresses a
					ddi(d, k)
				else:
					# suppresses b.c and b.d.e
					self.apply_delete(d[k], deletions[k])
					if not d[k]: # Suppress empty dicts
						ddi(d, k)

		return ReadOnlyDict(d) if self.freeze else d


	# Delete nested dict elements of any depth
	def apply_modify(self, d: Dict, modifications: Optional[Dict] = None) -> Dict:
		"""
		Method apply_modify:
		:param modifications: if None, self._nested_mods is used (set by constructor)
		modifications example: {'a':[<method>], 'b': {'c': [<method>], 'd': {'e': [<method>]}}}
		"""
		if self.unalterable:
			d = d.copy()

		dsi = dict.__setitem__

		for k, mod in (modifications or self._nested_mods).items():
			if k in d:
				if isinstance(mod, dict):
					dsi(d, k, self.apply_modify(d[k], mod))
				else:
					for mod in mod:
						dsi(d, k, mod(d[k]))

		return ReadOnlyDict(d) if self.freeze else d


	# Keep only dict elements of any depth
	######################################

	def apply_keep_only(self, d: Dict, nested_map: Optional[Dict] = None) -> Dict:
		"""
		Method apply_keep_only:
		:param nested_map: if None, self._nested_ko is used (set by constructor)
		modifications example: {'a':<set>, 'b': {'c': <set>, 'd': {'e': <set>}}}
		"""

		if self.unalterable:
			d = d.copy()

		dsi = dict.__setitem__
		for k, v in (nested_map or self._nested_ko).items():
			if k in d:
				if isinstance(v, dict):
					dsi(d, k, self.apply_keep_only(d[k], v))
				else: # should be a set
					dsi(d, k, {kk: vv for kk, vv in d[k].items() if kk in v})

		return ReadOnlyDict(d) if self.freeze else d


	def apply(self, d: Dict) -> Optional[Dict]:
		""" Modifies provided dict according to configured operations """
		try:
			for op in self.ops:
				d = op(d)
			return d
		except Exception as e:
			self.logger.error(
				f"Exception occured in {self.__class__.__name__}\n",
				exc_info=e
			)

		return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/db/FieldModel.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 19.10.2019
# Last Modified Date: 13.04.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from ampel.model.StrictModel import StrictModel

class FieldModel(StrictModel):

	field: str
	direction: int = 1

	def dict(self, **kwargs):
		return (self.field, self.direction)

	def get_id(self) -> str:
		"""
		Returns an indexId similar to what pymongo index_information outputs.
		Ex: [('tranId', 1), ('channel', 1)] -> tranId_1_channel_1
		"""
		# Shortcut
		return f"{self.field}_{self.direction}"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/mongo/update/MongoT3Ingester.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 30.05.2021
# Last Modified Date: 30.05.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from pymongo import UpdateOne
from typing import Dict, Any, Union
from ampel.enum.DocumentCode import DocumentCode
from ampel.content.T3Document import T3Document
from ampel.mongo.utils import maybe_use_each
from ampel.abstract.AbsDocIngester import AbsDocIngester


class MongoT3Ingester(AbsDocIngester[T3Document]):

	def ingest(self, doc: T3Document) -> None:
		raise NotImplementedError("Soon...")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/config/builder/ConfigValidator.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 03.09.2019
# Last Modified Date: 15.12.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Any
from ampel.base.BadConfig import BadConfig
from ampel.util.pretty import prettyjson
from ampel.model.ProcessModel import ProcessModel
from ampel.config.builder.BaseConfigChecker import BaseConfigChecker


class ConfigValidator(BaseConfigChecker):
    """
    Validate a config against the underlying model of process and the
    units defined in their configurations. No unit constructors are called.
    """

    def validate(self,
        ignore_inactive: bool = False,
        ignore_ressource_not_avail: bool = True,
        raise_exc: bool = True,
    ) -> dict[str, Any]:
        """
        :returns: config if check passed
        :raises: BadConfig
        """

        with self.loader.validate_unit_models():

            for tier, proc in self.iter_procs(ignore_inactive):
                config = self.config["process"][tier][proc]
                try:
                    ProcessModel(**config)
                except Exception as exc:
                    self.logger.debug(f"Invalid process config:\n{prettyjson(config)}")
                    self.logger.error(f"{tier} process {proc} from distribution {config.get('distrib')} is invalid", exc_info=exc)
                    raise BadConfig from exc

        return self.config



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/logging/LightLogRecord.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 22.04.2020
# Last Modified Date: 22.04.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from time import time
from typing import Union, List, Any, Dict, Optional
from ampel.types import ChannelId, StockId

class LightLogRecord:
	"""
	LogRecord class similar to the one provided by standard 'logging' module but faster.

	In []: %timeit LogRecord(name=12, pathname=None, level=12, lineno=12, exc_info=None, msg=None, args=None)
	Out[]: 4.7 µs ± 77.9 ns per loop (mean ± std. dev. of 7 runs, 100000 loops each)

	In []: %timeit LightLogRecord(name=12, levelno=1, msg=None)
	Out[]: 657 ns ± 6.51 ns per loop (mean ± std. dev. of 7 runs, 1000000 loops each)
	"""

	name: Union[int, str]
	levelno: int
	msg: Optional[Union[str, Dict[str, Any]]]
	channel: Optional[Union[ChannelId, List[ChannelId]]]
	stock: Optional[StockId]
	extra: Optional[Dict[str, Any]]

	def __init__(self,
		name: Union[int, str],
		levelno: int,
		msg: Optional[Union[str, Dict[str, Any]]] = None
	) -> None:

		d = self.__dict__
		d['name'] = name
		d['created'] = time()
		d['levelno'] = levelno
		d['msg'] = msg

	def getMessage(self):
		return self.msg

	def __getattr__(self, k):
		return self.__dict__[k] if k in self.__dict__ else None



from typing import Any

from ampel.abstract.AbsOpsUnit import AbsOpsUnit
from ampel.model.UnitModel import UnitModel
from ampel.abstract.AbsEventUnit import AbsEventUnit
from ampel.log import AmpelLogger, LogFlag, SHOUT
from ampel.log.utils import report_exception


class OpsProcessor(AbsEventUnit):

    execute: UnitModel
    update_beacon: bool = True
    log_profile: str = "console_verbose"

    def run(self) -> Any:

        logger = None

        try:
            beacon_col = self.context.db.get_collection("beacon")
            logger = AmpelLogger.from_profile(
                self.context,
                self.log_profile,
                base_flag=LogFlag.CORE | self.base_log_flag,
                force_refresh=True,
            )
            last_beacon = beacon_col.find_one({"_id": self.process_name})
            beacon = self.context.loader.new_context_unit(
                model=self.execute,
                context=self.context,
                sub_type=AbsOpsUnit,
                logger=logger,
            ).run(last_beacon)
            if beacon and self.update_beacon:
                last_beacon = beacon_col.update_one(
                    {"_id": self.process_name}, {"$set": beacon}, upsert=True
                )
        except Exception as e:

            if self.raise_exc:
                raise e

            if not logger:
                logger = AmpelLogger.get_logger()

            report_exception(
                self.context.db, logger, exc=e, info={"process": self.process_name}
            )

        finally:

            if not logger:
                logger = AmpelLogger.get_logger()

            # Feedback
            logger.log(SHOUT, f"Done running {self.process_name}")
            logger.flush()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/ingest/FilterModel.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 14.10.2019
# Last Modified Date: 20.05.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Optional, Dict, Literal
from ampel.model.UnitModel import UnitModel


class FilterModel(UnitModel):

	#: How to store rejection records
	reject: Optional[Dict[Literal['log', 'register'], UnitModel]]

	#: How to handle the defined filter if the associated stock is already in the database
	on_stock_match: Optional[Literal['bypass', 'overrule', 'silent_overrule']]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/config/collector/UnitConfigCollector.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 16.10.2019
# Last Modified Date: 10.10.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

import sys, re, importlib, traceback
from typing import List, Union, Dict, Any, Optional

from ampel.protocol.LoggingHandlerProtocol import AggregatingLoggingHandlerProtocol
from ampel.util.collections import ampel_iter
from ampel.model.StrictModel import StrictModel
from ampel.config.collector.ConfigCollector import ConfigCollector
from ampel.log import VERBOSE


class RemoteUnitDefinition(StrictModel):
	class_name: str
	base: List[str]


class UnitConfigCollector(ConfigCollector):


	def __init__(self, **kwargs) -> None:
		super().__init__(**kwargs)
		self.err_fqns: list[str] = []


	def add(self,
		arg: List[Union[Dict[str, str], str]],
		dist_name: str,
		version: Union[str, float, int],
		register_file: str
	) -> None:


		# Cosmetic
		if isinstance(self.logger.handlers[0], AggregatingLoggingHandlerProtocol):
			agg_int = self.logger.handlers[0].aggregate_interval
			self.logger.handlers[0].aggregate_interval = 1000

		# tolerate list containing only 1 element defined as dict
		for el in ampel_iter(arg):

			try:

				if isinstance(el, str):
					class_name = self.get_class_name(el)
					if not (mro := self.get_mro(el, class_name)):
						self.logger.break_aggregation()
						continue
					entry: Dict[str, Any] = {
						'fqn': el,
						'base': mro,
						'distrib': dist_name,
						'file': register_file,
						'version': version
					}

				elif isinstance(el, dict):
					try:
						d = RemoteUnitDefinition(**el)
					except Exception:
						self.error(
							'Unsupported unit definition (dict)' +
							self.distrib_hint(dist_name, register_file)
						)
						continue

					class_name = d.class_name
					entry = {
						'base': d.base,
						'distrib': dist_name,
						'file': register_file,
						'version': version
					}

				else:
					self.error(
						'Unsupported unit config format' +
						self.distrib_hint(dist_name, register_file)
					)
					continue

				if self.get(class_name):
					self.duplicated_entry(
						conf_key = class_name,
						section_detail = f'{self.tier} {self.conf_section}',
						new_file = register_file,
						new_dist = dist_name,
						prev_file = self.get(class_name).get('conf', 'unknown') # type: ignore
					)
					continue

				if "AmpelBaseModel" not in entry['base'] and "StrictModel" not in entry['base']:
					self.logger.info(
						f'Unrecognized base class for {self.conf_section} {class_name} ' +
						self.distrib_hint(dist_name, register_file)
					)
					continue

				for base in "AmpelABC", "AmpelBaseModel", "StrictModel":
					if base in entry["base"]:
						entry["base"].remove(base)

				if self.verbose:
					self.logger.log(VERBOSE,
						f'Adding {self.conf_section}: {class_name}'
					)

				self.__setitem__(class_name, entry)

			except Exception as e:

				if 'class_name' not in locals():
					class_name = 'Unknown'

				self.error(
					f'Error occured while loading {self.conf_section} {class_name} ' +
					self.distrib_hint(dist_name, register_file),
					exc_info=e
				)

		if isinstance(self.logger.handlers[0], AggregatingLoggingHandlerProtocol):
			self.logger.handlers[0].aggregate_interval = agg_int


	@staticmethod
	def get_class_name(fqn: str) -> str:
		"""
		Returns class name for a given fully qualified name
		Note: we here assume the ampel convention that module name equals class name
		(i.e that we have one class per module)
		"""
		return re.sub(r'.*\.', '', fqn) # noqa


	def get_mro(self, module_fqn: str, class_name: str) -> Optional[List[str]]:
		"""
		:param module_fqn: fully qualified name of module
		:param class_name: declared class name in the module specified by "module_fqn"
		:returns: the method resolution order (except for the first and last members
		that is of no use for our purpose) of the specified class
		"""

		try:
			UnitClass = getattr(
				# import using fully qualified name
				importlib.import_module(module_fqn),
				class_name
			)
		except Exception:

			self.err_fqns.append(module_fqn)
			# Suppress superfluous traceback entries
			print("", file=sys.stderr)
			print("Cannot import " + module_fqn, file=sys.stderr)
			print("-" * (len(module_fqn) + 14), file=sys.stderr)
			for el in traceback.format_exc().splitlines():
				if "_bootstrap" not in el and "in import_module" not in el:
					print(el, file=sys.stderr)

			return None

		return [
			el.__name__
			for el in UnitClass.__mro__[:-1]
			if 'ampel' in el.__module__
		]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/logging/handlers/ChanRecordBufHandler.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 01.05.2020
# Last Modified Date: 05.05.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Union, Optional, Dict, Any
from ampel.log.handlers.RecordBufferingHandler import RecordBufferingHandler
from ampel.protocol.LoggingHandlerProtocol import LoggingHandlerProtocol
from ampel.types import StockId, ChannelId


class ChanRecordBufHandler(RecordBufferingHandler):
	"""
	Record buferring handler that automatically associates
	every log record with a given channel when forwarding
	them to a target logger/handler.

	Note: the addition is not made by overriding the method handle(...)
	on purpose since it is unnecessary to morph/change all log entries:
	logs created during the rejection of alerts are usually not forwarded
	to a standard logger but rather forwarded to special handling classes
	(that save the information into a binary formatted file for instance)
	"""

	__slots__ = '_channel', '_extra'

	def __init__(self, level: int, channel: ChannelId, extra: Optional[Dict[str, Any]] = None) -> None:
		super().__init__(level)
		self._channel = channel
		self._extra = extra


	def forward(self,
		target: Union[LoggingHandlerProtocol],
		stock: Optional[StockId] = None,
		extra: Optional[Dict] = None,
		clear: bool = True
	) -> None:
		"""
		Forwards saved log records to provided logger/handler instance.
		Clears the internal record buffer.
		"""
		for rec in self.buffer:

			if rec.levelno >= target.level:

				rec.channel = self._channel # type: ignore[union-attr]

				if stock:
					rec.stock = stock # type: ignore[union-attr]

				if extra:
					if hasattr(rec, 'extra') and rec.extra: # type: ignore[union-attr]
						rec.extra |= extra # type: ignore[union-attr]
					else:
						rec.extra = extra # type: ignore[union-attr]

				if self._extra:
					if hasattr(rec, 'extra') and rec.extra: # type: ignore[union-attr]
						rec.extra |= self._extra # type: ignore[union-attr]
					else:
						rec.extra = self._extra # type: ignore[union-attr]

				target.handle(rec) # type: ignore

		if clear:
			self.buffer.clear()
			self.has_error = False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/t3/T3Processor.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 26.02.2018
# Last Modified Date: 17.12.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from importlib import import_module
from typing import Any, Optional, Annotated

from ampel.types import OneOrMany
from ampel.abstract.AbsEventUnit import AbsEventUnit
from ampel.abstract.AbsT3ControlUnit import AbsT3ControlUnit
from ampel.abstract.AbsProcessorTemplate import AbsProcessorTemplate
from ampel.view.T3Store import T3Store
from ampel.view.T3DocView import T3DocView
from ampel.model.UnitModel import UnitModel
from ampel.model.t3.T3IncludeDirective import T3IncludeDirective
from ampel.core.EventHandler import EventHandler
from ampel.abstract.AbsT3Supplier import AbsT3Supplier
from ampel.log import AmpelLogger, LogFlag, SHOUT


class T3Processor(AbsEventUnit):
	"""
	:param update_events: Record this event in the events collection
	:param extra_journal_tag:
	  tag(s) to add to the :class:`~ampel.content.JournalRecord.JournalRecord`
	  of each selected stock

	.. note:: by default, :func:`run` add entries to the journals of selected
	  stocks, store log records in the database, and records the invocation in
	  the events collection. To run an "anonymous" event that does not change
	  the state of the database:
	
	  - use a log_profile without db logging (profiles are defined under 'logging' in the ampel config)
	  - set update_events = False (no update to the events collection)
	  - set raise_exc = True (troubles collection will not be populated if an exception occurs)
	"""

	template: Optional[str] = None
	include: Optional[T3IncludeDirective]
	execute: OneOrMany[Annotated[UnitModel, AbsT3ControlUnit]]


	def __init__(self, **kwargs) -> None:

		if 'template' in kwargs:
			tpl_name = kwargs.pop("template")
			ctx = kwargs.pop("context")
			if ctx is None:
				raise ValueError("Context required")

			if kwargs['template'] not in ctx.config._config['template']:
				raise ValueError(f"Unknown process template: {tpl_name}")

			fqn = ctx.config._config['template'][tpl_name]
			class_name = fqn.split(".")[-1]
			Tpl = getattr(import_module(fqn), class_name)
			if not issubclass(Tpl, AbsProcessorTemplate):
				raise ValueError(f"Unexpected template type: {Tpl}")

			tpl = Tpl(**kwargs)
			kwargs = tpl.get_model(ctx.config._config, kwargs).dict()
			kwargs['context'] = ctx

		super().__init__(**kwargs)
		self.update_events = True


	def set_no_event(self) -> None:
		""" run t3 process \"silently\" """
		self.update_events = False


	def run(self) -> None:

		event_hdlr = None
		run_id = self.context.new_run_id()

		logger = AmpelLogger.from_profile(
			self.context, self.log_profile, run_id,
			base_flag = LogFlag.T3 | LogFlag.CORE | self.base_log_flag,
			force_refresh = True
		)

		# Create event doc
		event_hdlr = EventHandler(
			self.process_name, self.context.db, tier=3, run_id=run_id,
			raise_exc=self.raise_exc, dry_run=not self.update_events
		)

		try:

			# Feedback
			logger.log(SHOUT, f'Running {self.process_name}')

			loader = self.context.loader
			t3s = T3Store()

			if self.include:

				if self.include.docs:
					pass # later

				if (x := self.include.session):
					sdict: dict[str, Any] = {}
					for model in [x] if isinstance(x, UnitModel) else x:
						rd = loader.new_context_unit(
							model = model,
							context = self.context,
							sub_type = AbsT3Supplier,
							event_hdlr = event_hdlr,
							logger = logger
						).supply(t3s)
						if rd:
							sdict |= rd
					if sdict:
						t3s.add_session_info(sdict)

			units = self.context.config._config['unit']
			for i, um in enumerate([self.execute] if isinstance(self.execute, UnitModel) else self.execute):

				if um.unit not in units:
					raise ValueError(f"Unknown unit: {um.unit}")

				if "AbsT3ControlUnit" not in units[um.unit]['base']:
					raise ValueError("T3Processor executes only AbsT3ControlUnit units")

				t3_unit = loader.new_context_unit(
					model = um,
					context = self.context,
					sub_type = AbsT3ControlUnit,
					logger = logger,
					event_hdlr = event_hdlr,
					channel = self.channel
				)

				logger.info(
					f"Processing run block {i}",
					extra={'unit': t3_unit.__class__.__name__}
				)

				# Potential T3Document to be included in the t3 collection
				if (ret := t3_unit.process(t3s)):
					for t3d in ret:
						t3s.add_view(T3DocView.of(t3d, self.context.config))
						if 'meta' not in t3d:
							raise ValueError("Invalid T3Document")
						t3d['meta']['traceid'] = {'t3processor': self._trace_id}
						self.context.db.get_collection('t3').insert_one(t3d)


		except Exception as e:
			event_hdlr.handle_error(e, logger)

		# Update event document
		if event_hdlr:
			event_hdlr.add_extra(success=True)
			event_hdlr.update(logger)

		# Feedback
		logger.log(SHOUT, f'Done running {self.process_name}')
		logger.flush()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/mongo/query/var/events.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 11.07.2018
# Last Modified Date: 20.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from bson import ObjectId
from pymongo.collection import Collection
from datetime import datetime, timedelta
from typing import Dict, Optional, List, Literal, Any, Union, overload


def build_query(
	tier: Literal[0, 1, 2, 3] = 0,
	process_name: Optional[str] = None,
	lte_time: Optional[Union[dict, int, float]] = None,
	gte_time: Optional[Union[dict, int, float]] = None
) -> Dict[str, Any]:
	"""
	:param gte_time: unix timestamp or timedelta argument
	:param lte_time: unix timestamp or timedelta argument
	:returns: list of dict to be used as aggregation pipeline query parameters
	"""

	match: Dict[str, Any] = {}

	if gte_time:
		match['_id'] = {
			"$gte": ObjectId.from_datetime(_get_datetime(gte_time))
		}

	if lte_time:
		if '_id' not in match:
			match['_id'] = {}
		match['_id']['$lte'] = ObjectId.from_datetime(_get_datetime(lte_time))

	if tier:
		match['tier'] = tier

	if process_name:
		match['process'] = process_name

	return match


@overload
def get_last_run(
	col: Collection, process_name: str, require_success: bool,
	gte_time: Optional[Union[dict, float]], timestamp: Literal[True]
) -> Optional[Union[float]]:
	...

@overload
def get_last_run(
	col: Collection, process_name: str, require_success: bool,
	gte_time: Optional[Union[dict, float]], timestamp: Literal[False]
) -> Optional[Union[ObjectId]]:
	...

def get_last_run(
	col: Collection, process_name: str,
	require_success: bool,
	gte_time: Optional[Union[dict, float]] = None,
	timestamp: bool = True
) -> Optional[Union[float, ObjectId]]:
	"""
	:param gte_time: unix timestamp or timedelta argument
	"""

	query = build_query(tier=3, process_name=process_name, gte_time=gte_time)
	if require_success:
		query['success'] = True
	if ret := list(col.find(query).sort('_id', -1).limit(2)):
		if require_success and ret:
			return ret[0]['_id'].generation_time.timestamp() if timestamp else ret[0]['_id']
		elif len(ret) > 1:
			return ret[1]['_id'].generation_time.timestamp() if ret[1]['_id'] else ret[1]['_id']
	return None


def build_t0_stats_query(
	gte_time: Optional[Union[dict, int, float]] = None,
	lte_time: Optional[Union[dict, int, float]] = None
) -> List[Dict]:
	"""
	:param gte_time: unix timestamp or timedelta argument
	:param lte_time: unix timestamp or timedelta argument
	"""

	ret = build_query(tier=0, gte_time=gte_time, lte_time=lte_time)
	ret['metrics.count.alerts'] = {"$exists": True}

	return [
		{'$match': ret},
		{
			"$group": {
				"_id": 1,
				"alerts": {
					"$sum": "$metrics.count.alerts"
				}
			}
		}
	]


def _get_datetime(t: Union[int, float, dict]) -> datetime:
	if isinstance(t, (int, float)):
		return datetime.fromtimestamp(t)
	elif isinstance(t, dict):
		return datetime.today() + timedelta(**t)
	return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : ampel/core/abstract/AbsTargetSource.py
# License           : BSD-3-Clause
# Author            : jvs
# Date              : 15.08.2018
# Last Modified Date: 15.09.2018
# Last Modified By  : jvs

from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod


class AbsTargetSource(AmpelABC, abstract=True):
	""" Provides target fields for follow-up searches """

	@abstractmethod
	async def get_targets(self):
		"""
		:yields: a tuple ((ra,dec), radius, (date_min, date_max), [channels])
		"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/time/TimeLastRunModel.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 29.09.2018
# Last Modified Date: 31.07.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Union, Dict, Optional, Literal
from datetime import datetime, timedelta
from ampel.model.StrictModel import StrictModel
from ampel.mongo.query.var.events import get_last_run
from ampel.core.AmpelDB import AmpelDB


class TimeLastRunModel(StrictModel):

	match_type: Literal['time_last_run']
	process_name: str
	fallback: Union[None, Dict] = {'days': -1}
	require_success: bool = True


	def get_timestamp(self, **kwargs) -> Optional[float]:
		if ts := self._query_events_col(kwargs['db'], self):
			return ts
		if self.fallback:
			return (datetime.today() + timedelta(**self.fallback)).timestamp()
		return None


	@staticmethod
	def _query_events_col(ampel_db: AmpelDB, model: 'TimeLastRunModel') -> Optional[float]:

		col = ampel_db.get_collection('events')

		# First query the last 10 days
		res = get_last_run(
			col, model.process_name, model.require_success,
			gte_time=(datetime.today() - timedelta(days=10)).timestamp(),
			timestamp = True,
		)

		# If nothing is found, try querying the entire collection (days_back=None)
		if res is None:
			res = get_last_run(col, model.process_name, model.require_success, gte_time=None, timestamp=True)
			if not res:
				return None

		return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/abstract/AbsT3Supplier.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 15.07.2021
# Last Modified Date: 13.12.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Generic
from ampel.types import Traceless, T
from ampel.view.T3Store import T3Store
from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.log.AmpelLogger import AmpelLogger
from ampel.core.ContextUnit import ContextUnit
from ampel.core.EventHandler import EventHandler


class AbsT3Supplier(Generic[T], AmpelABC, ContextUnit, abstract=True):
	"""
	Abstract class for T3 suppliers
	"""

	logger: Traceless[AmpelLogger]
	event_hdlr: Traceless[EventHandler]

	@abstractmethod
	def supply(self, t3s: T3Store) -> T:
		...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/cli/T3BufferTextExporter.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 25.03.2021
# Last Modified Date: 10.12.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from json import dumps
from io import TextIOWrapper
from itertools import islice
from datetime import datetime
from typing import Optional, Union, TextIO, Generator

from ampel.view.T3Store import T3Store
from ampel.log.AmpelLogger import AmpelLogger
from ampel.abstract.AbsT3Stager import AbsT3Stager
from ampel.abstract.AbsIdMapper import AbsIdMapper
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.content.T3Document import T3Document
from ampel.util.serialize import walk_and_encode
from ampel.util.pretty import prettyjson
from ampel.util.getch import getch as fgetch

dsi = dict.__setitem__
ufts = datetime.utcfromtimestamp


class T3BufferTextExporter(AbsT3Stager):
	"""
	Exports AmpelBuffer instances as BSON into file
	"""

	fd: Union[TextIOWrapper, TextIO]
	raise_exc: bool = True
	close_fd: bool = True
	update_journal: bool = False
	verbose: bool = True
	pretty: bool = False
	id_mapper: Optional[AbsIdMapper] = None
	human_times: bool = True
	chunk_size: int = 200
	getch: bool = False


	def stage(self,
		gen: Generator[AmpelBuffer, None, None],
		t3s: T3Store
	) -> Optional[Generator[T3Document, None, None]]:

		logger = AmpelLogger.get_logger()
		func = prettyjson if self.pretty else dumps

		# Shortcuts
		fd = self.fd
		id_mapper = self.id_mapper
		human_times = self.human_times
		verbose = self.verbose
		convert_timestamps = self.convert_timestamps
		first = True

		fd.write('[\n')

		try:

			while (data := list(islice(gen, self.chunk_size))):

				walk_and_encode(data)

				if first and len(data) > 0:

					data = iter(data) # type: ignore
					el = next(data) # type: ignore
					if id_mapper:
						el['id'] = id_mapper.to_ext_id(el['id'])
					if human_times:
						convert_timestamps(el)
					if verbose:
						logger.info(f"Writing content (id: {el['id']})") # type: ignore
					fd.write(func(el)) # type: ignore
					first = False

				if self.getch and fgetch():
					fd.write('\n]\n')
					logger.info("Abording")
					return None

				for el in data:

					fd.write(",\n")
					if id_mapper:
						el['id'] = id_mapper.to_ext_id(el['id'])
					if human_times:
						convert_timestamps(el)
					if verbose:
						logger.info(f"Writing content (id: {el['id']})") # type: ignore
					fd.write(func(el)) # type: ignore

					if self.getch and fgetch():
						fd.write('\n]\n')
						logger.info("Abording")
						return None

			fd.write('\n]\n')

		finally:

			if self.close_fd:
				fd.flush()
				fd.close()

		return None


	def convert_timestamps(self, ab: AmpelBuffer) -> None:

		try:

			if stock_doc := ab.get('stock'):
				for k in (ts := stock_doc['ts']):
					dsi( # type: ignore[misc]
						ts, k, {
							'tied': ufts(ts[k]['tied']).isoformat(),
							'upd': ufts(ts[k]['upd']).isoformat()
						}
					)
				for je in stock_doc['journal']:
					dsi(je, 'ts', ufts(je['ts']).isoformat()) # type: ignore[arg-type]

			if t1s := ab.get('t1'):
				for t1doc in t1s:
					dsi(t1doc, 'added', ufts(t1doc['_id'].generation_time).isoformat()) # type: ignore[arg-type, typeddict-item]
					for t1meta in t1doc['meta']:
						dsi(t1meta, 'ts', ufts(t1meta['ts']).isoformat()) # type: ignore[arg-type]

			if t2s := ab.get('t2'):
				for t2doc in t2s:
					for t2meta in t2doc['meta']:
						dsi(t2meta, 'ts', ufts(t2meta['ts']).isoformat()) # type: ignore[arg-type]

		except Exception:
			import traceback, sys
			traceback.print_exc()
			print(ab)
			sys.exit("Exception occured")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/ProcessModel.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 06.10.2019
# Last Modified Date: 04.03.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

import schedule as sched
from pydantic import validator
from typing import Sequence, Optional, Literal, Union
from ampel.model.StrictModel import StrictModel
from ampel.types import ChannelId
from ampel.model.UnitModel import UnitModel
from ampel.config.ScheduleEvaluator import ScheduleEvaluator


class ProcessModel(StrictModel):

	name: str
	version: Union[int, float, str]
	active: bool = True
	tier: Optional[Literal[0, 1, 2, 3]]
	schedule: Sequence[str]
	channel: Optional[Union[ChannelId, Sequence[ChannelId]]]
	distrib: Optional[str]
	source: Optional[str]
	isolate: bool = True
	multiplier: int = 1
	log: Optional[str]
	controller: UnitModel = UnitModel(unit='DefaultProcessController')
	processor: UnitModel


	@validator('schedule', pre=True, each_item=False)
	def _cast_to_list(cls, v):
		if isinstance(v, str):
			return [v]
		return v


	@validator('schedule', each_item=False)
	def _check_schedule_validity(cls, schedule):
		"""
		Safety check for "schedule" parameters
		"""
		evaluator = ScheduleEvaluator()
		for el in schedule:
			if el == "super":
				continue
			try:
				evaluator(sched.Scheduler(), el).do(lambda x: None)
			except Exception:
				raise ValueError("Incorrect 'schedule' parameter")

		return schedule



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/mongo/model/AmpelColModel.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 19.10.2019
# Last Modified Date: 11.11.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import List, Optional, Union
from ampel.mongo.model.IndexModel import IndexModel
from ampel.mongo.model.ShortIndexModel import ShortIndexModel
from ampel.model.StrictModel import StrictModel

class AmpelColModel(StrictModel):
	name: str
	indexes: Optional[List[Union[ShortIndexModel, IndexModel]]]
	args: dict = {}



import re
from functools import partial
from pathlib import Path
from typing import Any, Callable, Literal, Optional, Union

from pydantic import BaseModel, validator

from ampel.model.ChannelModel import ChannelModel
from ampel.model.job.ExpressionParser import ExpressionParser
from ampel.model.UnitModel import UnitModel
from ampel.util.recursion import walk_and_process_dict


class OutputParameterSource(BaseModel):
    default: Optional[str]
    path: Path


class OutputParameter(BaseModel):
    name: str
    value_from: OutputParameterSource

    def value(self) -> Optional[str]:
        try:
            return self.value_from.path.read_text()
        except FileNotFoundError:
            return self.value_from.default


class TaskOutputs(BaseModel):
    parameters: list[OutputParameter] = []


class InputArtifactHttpSource(BaseModel):
    url: str


class InputArtifact(BaseModel):
    name: str
    path: Path
    http: InputArtifactHttpSource

    def value(self) -> Optional[str]:
        try:
            return self.path.read_text()
        except FileNotFoundError:
            return None


class InputParameter(BaseModel):
    name: str
    value: str


class TaskInputs(BaseModel):
    parameters: list[InputParameter] = []
    artifacts: list[InputArtifact] = []


class TaskUnitModel(UnitModel):
    title: str = ""
    multiplier: int = 1
    inputs: TaskInputs = TaskInputs()
    outputs: TaskOutputs = TaskOutputs()

    @validator("title", pre=True)
    def populate_title(cls, v, values):
        return v or values["unit"]


class TemplateUnitModel(BaseModel):
    title: str = ""
    template: str
    config: dict[str, Any]
    multiplier: int = 1
    inputs: TaskInputs = TaskInputs()
    outputs: TaskOutputs = TaskOutputs()

    @validator("title", pre=True)
    def populate_title(cls, v, values):
        return v or values["template"]


class MongoOpts(BaseModel):
    reset: bool = False
    prefix: str = "Ampel"


class Parameter(BaseModel):
    name: str
    value: str


class JobModel(BaseModel):
    name: str
    requirements: list[str] = []
    channel: list[dict[str, Any]] = []
    alias: dict[Literal["t0", "t1", "t2", "t3"], Any] = {}
    parameters: list[Parameter] = []
    mongo: MongoOpts = MongoOpts()
    task: list[Union[TemplateUnitModel, TaskUnitModel]]

    @validator("channel", each_item=True)
    def get_channel(cls, v):
        # work around incompatible AmpelBaseModel validation
        if not isinstance(v, dict):
            raise TypeError("channel must be a dict")
        if not "channel" in v:
            v["channel"] = v.pop("name")
        ChannelModel(**v)
        return v

    @classmethod
    def _transform_item(cls, v: str, transform: Callable[[str], str]) -> str:
        chunks = []
        pos = 0
        for match in re.finditer(r"\{\{(.*)\}\}", v):
            if match.span()[0] > pos:
                chunks.append(v[pos : match.span()[0]])
            chunks.append(transform(match.groups()[0].strip()))
            pos = match.span()[1]
        if pos < len(v):
            chunks.append(v[pos : len(v)])
        return "".join(chunks)

    @classmethod
    def _transform_expressions_callback(cls, path, k, d, **kwargs):
        for k, v in d.items():
            if isinstance(v, str):
                d[k] = cls._transform_item(v, kwargs["transform"])
            elif isinstance(v, list):
                d[k] = [
                    cls._transform_item(vv, kwargs["transform"])
                    if isinstance(vv, str)
                    else vv
                    for vv in v
                ]

    @classmethod
    def transform_expressions(
        cls, task_dict: dict, transformation: Callable[[str], str]
    ) -> dict:
        """
        Replace any expressions of the form {{ expr }} with the result of
        transformation(expr)
        """
        walk_and_process_dict(
            task_dict,
            callback=cls._transform_expressions_callback,
            transform=transformation,
        )
        return task_dict

    def resolve_expressions(self, target: dict, task: Union[TaskUnitModel,TemplateUnitModel]) -> dict:
        """
        Resolve any expressions of the form {{ expr }} found in string values of
        the target dict

        Supported expressions:

        - job parameters, e.g. {{ job.parameters.param }} resolves to the
          job-level parameter "param"
        - task outputs, e.g. {{ task.step-0.outputs.parameters.token }} resolves
          to the contents of the output file declared as "token" by the task
          named "step-0"
        """

        context = {
            "job": {
                "parameters": {param.name: param.value for param in self.parameters}
            },
            "task": {
                task.title: {
                    "outputs": {
                        "parameters": {
                            spec.name: spec
                            for spec in task.outputs.parameters
                        }
                    }
                }
                for task in self.task
            },
            "inputs": {
                "parameters": {param.name: param.value for param in task.inputs.parameters},
                "artifacts": {spec.name: spec for spec in task.inputs.artifacts},
            }
        }
        return self.transform_expressions(
            target,
            transformation=partial(ExpressionParser.evaluate, context=context),
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/cli/T3BufferBinaryExporter.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 25.03.2021
# Last Modified Date: 10.12.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from bson import encode
from io import BufferedWriter
from typing import Optional, Union, BinaryIO, Generator

from ampel.view.T3Store import T3Store
from ampel.log.AmpelLogger import AmpelLogger
from ampel.abstract.AbsT3Stager import AbsT3Stager
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.content.T3Document import T3Document
from ampel.abstract.AbsIdMapper import AbsIdMapper


class T3BufferBinaryExporter(AbsT3Stager):
	"""
	Exports AmpelBuffer instances as JSON (with base64 encoded bytes values) into file
	"""

	fd: Union[BufferedWriter, BinaryIO]
	raise_exc: bool = True
	update_journal: bool = False
	close_fd: bool = True
	verbose: bool = True
	id_mapper: Optional[AbsIdMapper] = None


	def stage(self,
		gen: Generator[AmpelBuffer, None, None],
		t3s: T3Store
	) -> Optional[Generator[T3Document, None, None]]:

		logger = AmpelLogger.get_logger()

		# Shortcuts
		fd = self.fd
		id_mapper = self.id_mapper
		verbose = self.verbose

		for el in gen:
			if verbose:
				logger.info(f"Writing content (id: {el['id']})") # type: ignore[str-bytes-safe]
			if id_mapper:
				el['id'] = id_mapper.to_ext_id(el['id'])
			fd.write(encode(el))

		logger.info("Closing file descriptor")
		if self.close_fd:
			fd.flush()
			fd.close()

		return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/time/TimeConstraintModel.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 29.09.2018
# Last Modified Date: 06.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Union, Optional
from ampel.model.time.TimeDeltaModel import TimeDeltaModel
from ampel.model.time.TimeLastRunModel import TimeLastRunModel
from ampel.model.time.TimeStringModel import TimeStringModel
from ampel.model.time.UnixTimeModel import UnixTimeModel
from ampel.model.time.QueryTimeModel import QueryTimeModel
from ampel.model.StrictModel import StrictModel


class TimeConstraintModel(StrictModel):
	"""
	Examples::
	  
	  TimeConstraintModel(
	      **{
	          "after": {
	              "match_type": "time_delta",
	              "days": -1
	          },
	          "before": {
	              "match_type": "time_string",
	              "dateTimeStr": "21/11/06 16:30",
	              "dateTimeFormat": "%d/%m/%y %H:%M"
	          }
	      }
	  )

	::
	  
	  TimeConstraintModel(
	      **{
	          "after": {
	              "match_type": "time_last_run",
	              "name": "val_test"
	          },
	          "before": {
	              "match_type": "unix_time",
	              "value": 1531306299
	          }
	      }
	  )
	"""

	before: Optional[
		Union[
			TimeDeltaModel, TimeLastRunModel,
			TimeStringModel, UnixTimeModel
		]
	] = None

	after: Optional[
		Union[
			TimeDeltaModel, TimeLastRunModel,
			TimeStringModel, UnixTimeModel
		]
	] = None


	def get_query_model(self, **kwargs) -> Optional[QueryTimeModel]:
		"""
		Call this method with db=<instance of AmpelDB>
		if your time constraint is based on TimeLastRunModel
		"""

		if self.before is None and self.after is None:
			return None

		return QueryTimeModel(
			before = self.before.get_timestamp(**kwargs) if self.before else None,
			after = self.after.get_timestamp(**kwargs) if self.after else None
		)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/cli/AbsCoreCommand.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 18.03.2021
# Last Modified Date: 11.11.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

import os
import re
from typing import Sequence, Dict, Any, Optional, TypeVar, Type, Iterator, Tuple
from ampel.abstract.AbsSecretProvider import AbsSecretProvider
from ampel.core.AmpelDB import AmpelDB
from ampel.config.AmpelConfig import AmpelConfig
from ampel.secret.AmpelVault import AmpelVault
from ampel.abstract.AbsCLIOperation import AbsCLIOperation
from ampel.core.AmpelContext import AmpelContext
from ampel.core.UnitLoader import UnitLoader
from ampel.log.AmpelLogger import AmpelLogger
from ampel.util.mappings import set_by_path
from ampel.util.freeze import recursive_freeze

custom_conf_patter = re.compile(r"^--[\w-]*(?:\.[\w-]+)*.*$")

T = TypeVar("T", bound = AmpelContext)

class AbsCoreCommand(AbsCLIOperation, abstract=True):

	def load_config(self,
		config_path: str,
		unknown_args: Sequence[str],
		logger: Optional[AmpelLogger] = None,
		freeze: bool = True,
		env_var_prefix: Optional[str] = "AMPEL_CONFIG_",
	) -> AmpelConfig:

		ampel_conf = AmpelConfig.load(config_path, freeze=False)

		if logger is None:
			logger = AmpelLogger.get_logger()

		if env_var_prefix is not None:
			for var, value in os.environ.items():
				if var.startswith(env_var_prefix):
					k = var[len(env_var_prefix):]
					v = _maybe_int(value)
					logger.info(f"Setting config parameter '{k}' from environment")
					set_by_path(ampel_conf._config, k, v)

		for k, v in self.get_custom_args(unknown_args):
			if ampel_conf.get(".".join(k.split(".")[:-1])) is None:
				logger.info(f"Cannot set ampel config parameter '{k}' (parent key does not exist)")
				continue

			logger.info(f"Setting config parameter '{k}' value to: {v}")
			set_by_path(ampel_conf._config, k, v)

		if freeze:
			ampel_conf._config = recursive_freeze(ampel_conf._config)

		return ampel_conf


	def get_custom_args(self, customizations: Sequence[str]) -> Iterator[Tuple[str, Any]]:

		it = iter(customizations)
		for el in it:
			if custom_conf_patter.match(el):
				if "=" in el:
					s = el.split("=")
					k = s[0][2:]
					v = _maybe_int(s[1])
				else:
					k = el[2:]
					try:
						v = _maybe_int(next(it))
					except StopIteration:
						break
				yield k, v


	def get_vault(self, args: Dict[str, Any]) -> Optional[AmpelVault]:
		vault = None
		if args.get('secrets'):
			from ampel.secret.DictSecretProvider import DictSecretProvider
			from ampel.secret.DirSecretProvider import DirSecretProvider
			if os.path.isdir(args['secrets']):
				provider: AbsSecretProvider = DirSecretProvider(args['secrets'])
			else:
				provider = DictSecretProvider.load(args['secrets'])
			vault = AmpelVault([provider])
		return vault


	def get_db(self,
		config: AmpelConfig,
		vault: Optional[AmpelVault] = None,
		require_existing_db: bool = True,
		one_db: bool = False
	) -> AmpelDB:

		try:
			return AmpelDB.new(
				config,
				vault,
				require_exists = require_existing_db,
				one_db = one_db
			)
		except Exception as e:
			if "Databases with prefix" in str(e):
				s = "Databases with prefix " + config.get('mongo.prefix', str, raise_exc=True) + " do not exist"
				raise SystemExit("\n" + "="*len(s) + "\n" + s + "\n" + "="*len(s) + "\n")
			raise e


	def get_context(self,
		args: Dict[str, Any],
		unknown_args: Sequence[str],
		logger: Optional[AmpelLogger] = None,
		freeze_config: bool = True,
		ContextClass: Type[T] = AmpelContext, # type: ignore[assignment]
		require_existing_db: bool = True,
		one_db: bool = False,
		**kwargs
	) -> T:

		if logger is None:
			logger = AmpelLogger.get_logger()

		config = self.load_config(
			args['config'], unknown_args, logger, freeze = freeze_config
		)

		vault = self.get_vault(args)
		db = self.get_db(config, vault, require_existing_db, one_db)
		return ContextClass(
			config = config,
			db = db,
			loader = UnitLoader(config, db=db, vault=vault),
			**kwargs
		)


	def convert_logical_args(self, name: str, args: Dict[str, Any]) -> None:

		for k in (f"with_{name}", f"with_{name}s_and", f"with_{name}s_or"):
			if args.get(k):
				if name not in args:
					args[name] = {}
				args[name]['with'] = args[k]
				break

		for k in (f"without_{name}", f"without_{name}s_and", f"without_{name}s_or"):
			if args.get(k):
				if name not in args:
					args[name] = {}
				args[name]['without'] = args[k]
				break


def _maybe_int(stringy):
	try:
		return int(stringy)
	except Exception:
		return stringy



import os
from typing import Any, ClassVar, List, Optional

from prometheus_client import CollectorRegistry, Counter, Gauge, Histogram
from prometheus_client.multiprocess import MultiProcessCollector


class AmpelMetricsRegistry:

    _registry: ClassVar[Optional[CollectorRegistry]] = None
    _standalone_collectors: ClassVar[List[Any]] = []

    @classmethod
    def registry(cls) -> CollectorRegistry:
        if cls._registry is None:
            cls._registry = CollectorRegistry()
        return cls._registry

    @classmethod
    def collect(cls):
        if "PROMETHEUS_MULTIPROC_DIR" in os.environ:
            yield from MultiProcessCollector(None).collect()
        else:
            yield from cls.registry().collect()
        for collector in cls._standalone_collectors:
            yield from collector.collect()

    @classmethod
    def register_collector(cls, collector):
        cls._standalone_collectors.append(collector)

    @classmethod
    def deregister_collector(cls, collector):
        cls._standalone_collectors.remove(collector)

    @classmethod
    def counter(
        cls, name, documentation, unit="", labelnames=(), subsystem=""
    ) -> Counter:
        return Counter(
            name,
            documentation,
            unit=unit,
            labelnames=labelnames,
            namespace="ampel",
            subsystem=subsystem,
            registry=cls.registry(),
        )

    @classmethod
    def gauge(
        cls,
        name,
        documentation,
        unit="",
        labelnames=(),
        subsystem="",
        multiprocess_mode="livesum",
    ) -> Gauge:
        return Gauge(
            name,
            documentation,
            unit=unit,
            labelnames=labelnames,
            namespace="ampel",
            subsystem=subsystem,
            registry=cls.registry(),
            multiprocess_mode=multiprocess_mode,
        )

    @classmethod
    def histogram(
        cls,
        name,
        documentation,
        unit="",
        labelnames=(),
        subsystem="",
        buckets=Histogram.DEFAULT_BUCKETS,
    ) -> Histogram:
        return Histogram(
            name,
            documentation,
            unit=unit,
            labelnames=labelnames,
            namespace="ampel",
            subsystem=subsystem,
            registry=cls.registry(),
            buckets=buckets,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/config/cli.py
# License           : BSD-3-Clause
# Author            : Jakob van Santen <jakob.van.santen@desy.de>
# Date              : 26.08.2020
# Last Modified Date: 26.08.2020
# Last Modified By  : Jakob van Santen <jakob.van.santen@desy.de>

import json, subprocess, sys, yaml
from io import StringIO
from typing import Any, Dict, Iterable, Mapping, Optional, TextIO
from argparse import ArgumentParser, ArgumentTypeError, FileType, Namespace

from ampel.base.BadConfig import BadConfig
from ampel.config.builder.DistConfigBuilder import DistConfigBuilder
from ampel.log.utils import log_exception
from ampel.core.AmpelContext import AmpelContext
from ampel.secret.AmpelVault import AmpelVault
from ampel.secret.DictSecretProvider import DictSecretProvider
from ampel.secret.PotemkinSecretProvider import PotemkinSecretProvider


def transform(args: Namespace) -> None:
    """Transform existing configuration with jq"""
    try:
        with FileType()(args.filter) as f:
            jq_args = [f.read()]
    except ArgumentTypeError:
        jq_args = [args.filter]
    # Use a custom transformation to losslessly round-trip from YAML to JSON,
    # in particular:
    # - wrap large ints to prevent truncation to double precision
    # - preserve non-string keys
    input_json = json.dumps(_to_json(yaml.safe_load(args.config_file)))
    config = json.loads(
        subprocess.check_output(["jq"] + jq_args, input=input_json.encode()),
        object_hook=_from_json,
    )
    with StringIO() as output_yaml:
        yaml.dump(config, output_yaml, sort_keys=False)
        if args.validate:
            output_yaml.seek(0)
            _validate(output_yaml)
        output_yaml.seek(0)
        args.output_file.write(output_yaml.read())


def build(args: Namespace) -> int:
    """Build config file from installed distributions"""
    cb = DistConfigBuilder(verbose=args.verbose)
    try:
        cb.load_distributions()
        config = cb.build_config(
            stop_on_errors=0 if args.ignore_errors else 2,
            config_validator="ConfigValidator",
            get_unit_env=args.get_env,
        )
    except Exception as exc:
        # assume that BadConfig means the error was already logged
        if not isinstance(exc, BadConfig):
            log_exception(cb.logger, exc)
        return 1
    yaml.dump(
        config, args.output_file if args.output_file else sys.stdout, sort_keys=False
    )
    return 0


def _load_dict(source: TextIO) -> Dict[str, Any]:
    if isinstance((payload := yaml.safe_load(source)), dict):
        return payload
    else:
        raise TypeError("buf does not deserialize to a dict")


def _validate(config_file: TextIO, secrets: Optional[TextIO] = None) -> None:
    from ampel.model.ChannelModel import ChannelModel
    from ampel.model.ProcessModel import ProcessModel

    ctx = AmpelContext.load(
        _load_dict(config_file),
        secrets=AmpelVault(providers=[(
            DictSecretProvider(_load_dict(secrets))
            if secrets is not None
            else PotemkinSecretProvider()
        )]),
    )
    with ctx.loader.validate_unit_models():
        for channel in ctx.config.get(
            "channel", Dict[str, Any], raise_exc=True
        ).values():
            ChannelModel(**{k: v for k, v in channel.items() if k not in {"template"}})
        for tier in range(3):
            for process in ctx.config.get(
                f"process.t{tier}", Dict[str, Any], raise_exc=True
            ).values():
                ProcessModel(**process)


def validate(args: Namespace) -> None:
    """Validate configuration"""
    _validate(args.config_file, args.secrets)


def main():
    parser = ArgumentParser()

    subparsers = parser.add_subparsers(help="command help", dest="command")
    subparsers.required = True

    def add_command(f, name=None):
        if name is None:
            name = f.__name__
        p = subparsers.add_parser(name, help=f.__doc__)
        p.set_defaults(func=f)
        return p

    p = add_command(build)
    p.add_argument("-v", "--verbose", default=False, action="store_true")
    p.add_argument("--ignore-errors", default=False, action="store_true")
    p.add_argument("--no-get-env", dest="get_env", default=True, action="store_false",
        help="Skip Python dependency detection when gathering units."
    )
    p.add_argument("-o", "--output-file", type=FileType("w"))

    p = add_command(validate)
    p.add_argument("config_file", type=FileType("r"))
    p.add_argument("--secrets", type=FileType("r"))

    p = add_command(transform)
    p.add_argument("filter", help="jq filter (or path to file containing one)")
    p.add_argument("config_file", type=FileType("r"))
    p.add_argument(
        "--validate",
        default=False,
        action="store_true",
        help="Validate config after applying transformation",
    )
    p.add_argument("-o", "--output-file", type=FileType("w"), default=sys.stdout)

    args = parser.parse_args()
    sys.exit(args.func(args))


def _to_json(obj):
    """Get JSON-compliant representation of obj"""
    if isinstance(obj, Mapping):
        assert "__nonstring_keys" not in obj
        doc = {str(k): _to_json(v) for k, v in obj.items()}
        nonstring_keys = {
            str(k): _to_json(k) for k in obj.keys() if not isinstance(k, str)
        }
        if nonstring_keys:
            doc["__nonstring_keys"] = nonstring_keys
        return doc
    elif isinstance(obj, Iterable) and not isinstance(obj, str):
        return [_to_json(v) for v in obj]
    elif isinstance(obj, int) and abs(obj) >> 53:
        # use canonical BSON representation for ints larger than the precision
        # of a double
        return {"$numberLong": str(obj)}
    else:
        return obj


def _from_json(doc):
    """Invert _to_json()"""
    if "$numberLong" in doc:
        return int(doc["$numberLong"])
    elif "__nonstring_keys" in doc:
        nonstring_keys = doc.pop("__nonstring_keys")
        return {nonstring_keys[k]: v for k, v in doc.items()}
    else:
        return doc



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/abstract/AbsDocIngester.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 10.03.2020
# Last Modified Date: 09.10.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, Optional, Any, Generic
from ampel.types import T
from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.mongo.update.DBUpdatesBuffer import DBUpdatesBuffer


class AbsDocIngester(Generic[T], AmpelABC, AmpelBaseModel, abstract=True):

	updates_buffer: DBUpdatesBuffer

	@abstractmethod
	def ingest(self, doc: T) -> None:
		...

	def get_stats(self, reset: bool = True) -> Optional[Dict[str, Any]]:
		return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/secret/AEAbsSecretProvider.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 20.06.2021
# Last Modified Date: 20.06.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from sjcl import SJCL
from typing import Type, Union, Iterable
from ampel.abstract.AbsSecretProvider import AbsSecretProvider
from ampel.secret.AESecret import AESecret
from ampel.secret.Secret import Secret


class AESecretProvider(AbsSecretProvider):

	def __init__(self, pwds: Union[str, Iterable[str]]):
		self.sjcl = SJCL()
		self.pwds = [pwds] if isinstance(pwds, str) else pwds


	def tell(self, arg: Secret, ValueType: Type) -> bool:
		"""
		Potentially update an initialized Secret instance with
		the actual sensitive information associable with it.
		:returns: True if the Secret was told/resolved or
		False if the provided Secret is unknown to this secret provider
		"""

		if not issubclass(str, ValueType):
			return False

		if isinstance(arg, AESecret):
			for pwd in self.pwds:
				try:
					arg.set(
						self.sjcl \
							.decrypt(arg.dict(), pwd) \
							.decode("utf-8")
					)
					return True
				except Exception:
					continue

		return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/log/__init__.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 11.06.2020
# Last Modified Date: 11.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

# flake8: noqa: F401
from .AmpelLogger import AmpelLogger, DEBUG, INFO, VERBOSE, SHOUT, WARNING, ERROR
from .LogFlag import LogFlag
